# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# WAMP message type IDs organized by message categories:
# - Category 1: Neither Payload nor Forwarding (12 messages)
# - Category 2: Payload Only (0 messages) - architecturally empty
# - Category 3: Forwarding Only (3 messages)
# - Category 4: Both Payload and Forwarding (7 messages)
class MessageType(object):
    # No valid message type
    NULL = 0
    # [Category 1] Session opening, authentication and closing
    HELLO = 1
    WELCOME = 2
    ABORT = 3
    CHALLENGE = 4
    AUTHENTICATE = 5
    GOODBYE = 6
    # [Category 4] Used in both PubSub and RPC
    ERROR = 8
    # Publish & Subscribe (PubSub)
    PUBLISH = 16
    PUBLISHED = 17
    SUBSCRIBE = 32
    SUBSCRIBED = 33
    UNSUBSCRIBE = 34
    UNSUBSCRIBED = 35
    EVENT = 36
    EVENT_RECEIVED = 37
    # Remote Procedure Calls (RPC)
    CALL = 48
    CANCEL = 49
    RESULT = 50
    REGISTER = 64
    REGISTERED = 65
    UNREGISTER = 66
    UNREGISTERED = 67
    INVOCATION = 68
    INTERRUPT = 69
    YIELD = 70



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# Remote procedure invocation policy, for use with shared registrations.
class InvocationPolicy(object):
    # Standard invocation policy: a procedure may only be registered by at most
    # one callee at a time. This is the default.
    SINGLE = 0
    # First callee invocation policy: all calls to a procedure are forwarded to
    # the callee that first registered. Should that callee unregister, calls
    # are forwarded to the callee that registered next after the former.
    FIRST = 1
    # Last callee invocation policy: all calls to a procedure are forwwrded to
    # the callee that last registered. Should that callee unregister, calls
    # are forwarded to the callee that registered next before the former.
    LAST = 2
    # Round-robin invocation policy: calls are forward to all registered callees
    # in a round-robin fashion. Callees that unregister are removed from the
    # round-robin list, and new callees registering will get calls immediately
    # in the next dispatching round.
    ROUNDROBIN = 3
    # Random invocation policy: calls are forwarded to all registered callees
    # in a random fashion. Callees that unregister are removed from the set
    # of callees randomly drawn from to dispatch incoming calls.
    RANDOM = 4



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import os
import struct

if os.environ.get("USE_TWISTED", False):
    from base64 import b64decode
    from unittest.mock import MagicMock, patch

    from txaio.testutil import replace_loop

    from autobahn.testutil import FakeTransport
    from autobahn.twisted.websocket import (
        WebSocketClientFactory,
        WebSocketClientProtocol,
        WebSocketServerFactory,
        WebSocketServerProtocol,
    )
    from autobahn.websocket.compress_deflate import PerMessageDeflate
    from twisted.internet.address import IPv4Address
    from twisted.internet.task import Clock
    from twisted.trial import unittest

    @patch("base64.b64encode")
    def create_client_frame(b64patch, **kwargs):
        """
        Kind-of hack-y; maybe better to re-factor the Protocol to have a
        frame-encoder method-call? Anyway, makes a throwaway protocol
        encode a frame for us, collects the .sendData call and returns
        the data that would have gone out. Accepts all the kwargs that
        WebSocketClientProtocol.sendFrame() accepts.
        """

        # only real way to inject a "known" secret-key for the headers
        # to line up... :/
        b64patch.return_value = b"QIatSt9QkZPyS4QQfdufO8TgkL0="

        factory = WebSocketClientFactory(protocols=["wamp.2.json"])
        factory.protocol = WebSocketClientProtocol
        factory.doStart()
        proto = factory.buildProtocol(IPv4Address("TCP", "127.0.0.9", 65534))
        proto.transport = FakeTransport()
        proto.connectionMade()
        proto.data = mock_handshake_server
        proto.processHandshake()

        data = []

        def collect(d, *args):
            data.append(d)

        proto.sendData = collect

        proto.sendFrame(**kwargs)
        return b"".join(data)

    # beware the evils of line-endings...
    mock_handshake_client = b"GET / HTTP/1.1\r\nUser-Agent: AutobahnPython/0.10.2\r\nHost: localhost:80\r\nUpgrade: WebSocket\r\nConnection: Upgrade\r\nPragma: no-cache\r\nCache-Control: no-cache\r\nSec-WebSocket-Key: 6Jid6RgXpH0RVegaNSs/4g==\r\nSec-WebSocket-Protocol: wamp.2.json\r\nSec-WebSocket-Version: 13\r\n\r\n"

    mock_handshake_server = b'HTTP/1.1 101 Switching Protocols\r\nServer: AutobahnPython/0.10.2\r\nX-Powered-By: AutobahnPython/0.10.2\r\nUpgrade: WebSocket\r\nConnection: Upgrade\r\nSec-WebSocket-Protocol: wamp.2.json\r\nSec-WebSocket-Accept: QIatSt9QkZPyS4QQfdufO8TgkL0=\r\n\r\n\x81~\x02\x19[1,"crossbar",{"roles":{"subscriber":{"features":{"publisher_identification":true,"pattern_based_subscription":true,"subscription_revocation":true}},"publisher":{"features":{"publisher_identification":true,"publisher_exclusion":true,"subscriber_blackwhite_listing":true}},"caller":{"features":{"caller_identification":true,"progressive_call_results":true}},"callee":{"features":{"progressive_call_results":true,"pattern_based_registration":true,"registration_revocation":true,"shared_registration":true,"caller_identification":true}}}}]\x18'

    class TestDeflate(unittest.TestCase):
        def test_max_size(self):
            decoder = PerMessageDeflate(
                is_server=False,
                server_no_context_takeover=False,
                client_no_context_takeover=False,
                server_max_window_bits=15,
                client_max_window_bits=15,
                mem_level=8,
                max_message_size=10,
            )

            # 2000 'x' characters compressed
            compressed_data = (
                b"\xab\xa8\x18\x05\xa3`\x14\x8c\x82Q0\nF\xc1P\x07\x00\xcf@\xa9\xae"
            )

            decoder.start_decompress_message()
            data = decoder.decompress_message_data(compressed_data)

            # since we set max_message_size, we should only get that
            # many bytes back.
            self.assertEqual(data, b"x" * 10)

        def test_no_max_size(self):
            decoder = PerMessageDeflate(
                is_server=False,
                server_no_context_takeover=False,
                client_no_context_takeover=False,
                server_max_window_bits=15,
                client_max_window_bits=15,
                mem_level=None,
            )

            # 2000 'x' characters compressed
            compressed_data = (
                b"\xab\xa8\x18\x05\xa3`\x14\x8c\x82Q0\nF\xc1P\x07\x00\xcf@\xa9\xae"
            )

            decoder.start_decompress_message()
            data = decoder.decompress_message_data(compressed_data)

            self.assertEqual(data, b"x" * 2000)

    class TestClient(unittest.TestCase):
        def setUp(self):
            self.factory = WebSocketClientFactory(protocols=["wamp.2.json"])
            self.factory.protocol = WebSocketClientProtocol
            self.factory.doStart()

            self.proto = self.factory.buildProtocol(
                IPv4Address("TCP", "127.0.0.1", 65534)
            )
            self.transport = FakeTransport()
            self.proto.transport = self.transport
            self.proto.connectionMade()

        def tearDown(self):
            if self.proto.openHandshakeTimeoutCall:
                self.proto.openHandshakeTimeoutCall.cancel()
            self.factory.doStop()
            # not really necessary, but ...
            del self.factory
            del self.proto

        def test_missing_reason_raw(self):
            # we want to hit the "STATE_OPEN" case, so pretend we're there
            self.proto.echoCloseCodeReason = True
            self.proto.state = self.proto.STATE_OPEN
            self.proto.websocket_version = 1

            self.proto.sendCloseFrame = MagicMock()

            self.proto.onCloseFrame(1000, None)

        def test_unclean_timeout_client(self):
            """
            make a delayed call to drop the connection (client-side)
            """

            if False:
                self.proto.factory._log = print

            # get to STATE_OPEN
            self.proto.websocket_key = b64decode("6Jid6RgXpH0RVegaNSs/4g==")
            self.proto.data = mock_handshake_server
            self.proto.processHandshake()
            self.assertEqual(self.proto.state, WebSocketServerProtocol.STATE_OPEN)
            self.assertTrue(self.proto.serverConnectionDropTimeout > 0)

            with replace_loop(Clock()) as reactor:
                # now 'do the test' and transition to CLOSING
                self.proto.sendCloseFrame()
                self.proto.onCloseFrame(1000, b"raw reason")

                # check we scheduled a call
                self.assertEqual(len(reactor.calls), 1)
                self.assertEqual(
                    reactor.calls[0].func, self.proto.onServerConnectionDropTimeout
                )
                self.assertEqual(
                    reactor.calls[0].getTime(), self.proto.serverConnectionDropTimeout
                )

                # now, advance the clock past the call (and thereby
                # execute it)
                reactor.advance(self.proto.closeHandshakeTimeout + 1)

                # we should have called abortConnection
                self.assertTrue(self.proto.transport.abort_called())
                # ...too "internal" for an assert?
                self.assertEqual(self.proto.state, WebSocketServerProtocol.STATE_CLOSED)

    class TestPing(unittest.TestCase):
        def setUp(self):
            self.factory = WebSocketServerFactory(protocols=["wamp.2.json"])
            self.factory.protocol = WebSocketServerProtocol
            self.factory.doStart()

            self.proto = self.factory.buildProtocol(
                IPv4Address("TCP", "127.0.0.1", 65534)
            )
            self.transport = MagicMock()
            self.proto.transport = self.transport
            self.proto.connectionMade()

        def tearDown(self):
            if self.proto.openHandshakeTimeoutCall:
                self.proto.openHandshakeTimeoutCall.cancel()
            self.factory.doStop()
            # not really necessary, but ...
            del self.factory
            del self.proto

        def test_unclean_timeout(self):
            """
            make a delayed call to drop the connection
            """
            # first we have to drive the protocol to STATE_CLOSING
            # ... which we achieve by sendCloseFrame after we're in
            # STATE_OPEN
            # XXX double-check this is the correct code-path to get here
            # "normally"?

            # get to STATE_OPEN
            self.proto.data = mock_handshake_client
            self.proto.processHandshake()
            self.assertTrue(self.proto.state == WebSocketServerProtocol.STATE_OPEN)

            with replace_loop(Clock()) as reactor:
                # now 'do the test' and transition to CLOSING
                self.proto.sendCloseFrame()

                # check we scheduled a call
                self.assertEqual(len(reactor.calls), 1)

                # now, advance the clock past the call (and thereby
                # execute it)
                reactor.advance(self.proto.closeHandshakeTimeout + 1)

                # we should have called abortConnection
                self.assertEqual(
                    "call.abortConnection()", str(self.proto.transport.method_calls[-1])
                )
                self.assertTrue(self.proto.transport.abortConnection.called)
                # ...too "internal" for an assert?
                self.assertEqual(self.proto.state, WebSocketServerProtocol.STATE_CLOSED)

        def test_auto_pingpong_timeout(self):
            """
            autoping and autoping-timeout timing
            """
            # options are evaluated in succeedHandshake, called below
            self.proto.autoPingInterval = 5
            self.proto.autoPingTimeout = 2

            with replace_loop(Clock()) as reactor:
                # get to STATE_OPEN
                self.proto.data = mock_handshake_client
                self.proto.processHandshake()
                self.assertTrue(self.proto.state == WebSocketServerProtocol.STATE_OPEN)

                # we should have scheduled an autoPing
                self.assertEqual(1, len(reactor.calls))

                # advance past first auto-ping timeout
                reactor.advance(5)

                # first element from args tuple from transport.write()
                # call is our data
                self.assertTrue(self.transport.write.called)
                data = self.transport.write.call_args[0][0]

                _data = bytes([data[0]])

                # the opcode is the lower 7 bits of the first byte.
                (opcode,) = struct.unpack("B", _data)
                opcode = opcode & (~0x80)

                # ... and should be "9" for ping
                self.assertEqual(9, opcode)

                # Because we have autoPingTimeout there should be
                # another delayed-called created now
                self.assertEqual(1, len(reactor.calls))
                self.assertNotEqual(self.proto.state, self.proto.STATE_CLOSED)

                # ...which we'll now cause to trigger, aborting the connection
                reactor.advance(3)
                self.assertEqual(self.proto.state, self.proto.STATE_CLOSED)

        def test_auto_ping_got_pong(self):
            """
            auto-ping with correct reply cancels timeout
            """
            # options are evaluated in succeedHandshake, called below
            self.proto.autoPingInterval = 5
            self.proto.autoPingTimeout = 2

            with replace_loop(Clock()) as reactor:
                # get to STATE_OPEN
                self.proto.data = mock_handshake_client
                self.proto.processHandshake()
                self.assertTrue(self.proto.state == WebSocketServerProtocol.STATE_OPEN)

                # we should have scheduled an autoPing
                self.assertEqual(1, len(reactor.calls))

                # advance past first auto-ping timeout
                reactor.advance(5)

                # should have an auto-ping timeout scheduled, and we
                # save it for later (to check it got cancelled)
                self.assertEqual(1, len(reactor.calls))
                timeout_call = reactor.calls[0]

                # elsewhere we check that we actually send an opcode-9
                # message; now we just blindly inject our own reply
                # with a PONG frame

                frame = create_client_frame(
                    opcode=10, payload=self.proto.autoPingPending
                )
                self.proto.data = frame
                # really needed twice; does header first, then rest
                self.proto.processData()
                self.proto.processData()

                # which should have cancelled the call
                self.assertTrue(timeout_call.cancelled)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# WAMP authentication method (see: WAMP protocol spec, "13. Authentication Methods").
class AuthMethod(object):
    # Not set / not authenticated: `anonymous`.
    NULL = 0
    # WAMP Ticket authentication: `ticket` (see: WAMP protocol spec, "13.1. Ticket-based Authentication").
    TICKET = 1
    # WAMP Challenge-Response authentication: `wampcra` (see: WAMP protocol spec, "13.2. Challenge Response Authentication").
    CRA = 2
    # WAMP Salted Challenge Response authentication: `wamp-scram` (see: WAMP protocol spec, "13.3. Salted Challenge Response Authentication").
    SCRAM = 3
    # WAMP Cryptosign authentication: `cryptosign` (see: WAMP protocol spec, "13.4. Cryptosign-based Authentication").
    CRYPTOSIGN = 4



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################


class FakeTransport(object):
    _written = b""
    _open = True

    def __init__(self):
        self._abort_calls = []

    def abortConnection(self, *args, **kw):
        self._abort_calls.append((args, kw))

    def write(self, msg):
        if not self._open:
            raise Exception("Can't write to a closed connection")
        self._written = self._written + msg

    def loseConnection(self):
        self._open = False

    def registerProducer(self, producer, streaming):
        # https://twistedmatrix.com/documents/current/api/twisted.internet.interfaces.IConsumer.html
        raise NotImplementedError

    def unregisterProducer(self):
        # do nothing is correct! until we fake implement registerProducer ..;)
        pass

    def getPeer(self):
        # for Twisted, this would be an IAddress
        class _FakePeer(object):
            pass

        return _FakePeer()

    def getHost(self):
        # for Twisted, this would be an IAddress
        class _FakeHost(object):
            pass

        return _FakeHost()

    def abort_called(self):
        return len(self._abort_calls) > 0



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################


from zope.interface import implementer

from twisted.protocols.policies import ProtocolWrapper

try:
    # starting from Twisted 22.10.0 we have `notFound`
    from twisted.web.pages import notFound
except ImportError:
    try:
        # In Twisted < 22.10.0 && > 12.2 this was called `NoResource`
        from twisted.web.resource import NoResource as notFound
    except ImportError:
        # And in Twisted < 12.2 this was in a different place
        from twisted.web.error import NoResource as notFound

from twisted.web.resource import IResource, Resource

# The following triggers an import of reactor at module level!
#
from twisted.web.server import NOT_DONE_YET

__all__ = (
    "WSGIRootResource",
    "WebSocketResource",
)


class WSGIRootResource(Resource):
    """
    Root resource when you want a WSGI resource be the default serving
    resource for a Twisted Web site, but have subpaths served by
    different resources.

    This is a hack needed since
    `twisted.web.wsgi.WSGIResource <http://twistedmatrix.com/documents/current/api/twisted.web.wsgi.WSGIResource.html>`_.
    does not provide a ``putChild()`` method.

    .. seealso::

       * `Autobahn Twisted Web WSGI example <https://github.com/crossbario/autobahn-python/tree/master/examples/twisted/websocket/echo_wsgi>`_
       * `Original hack <http://blog.vrplumber.com/index.php?/archives/2426-Making-your-Twisted-resources-a-url-sub-tree-of-your-WSGI-resource....html>`_
    """

    def __init__(self, wsgiResource, children):
        """

        :param wsgiResource: The WSGI to serve as root resource.
        :type wsgiResource: Instance of `twisted.web.wsgi.WSGIResource <http://twistedmatrix.com/documents/current/api/twisted.web.wsgi.WSGIResource.html>`_.
        :param children: A dictionary with string keys constituting URL subpaths, and Twisted Web resources as values.
        :type children: dict
        """
        Resource.__init__(self)
        self._wsgiResource = wsgiResource
        self.children = children

    def getChild(self, path, request):
        request.prepath.pop()
        request.postpath.insert(0, path)
        return self._wsgiResource


@implementer(IResource)
class WebSocketResource(object):
    """
    A Twisted Web resource for WebSocket.
    """

    isLeaf = True

    def __init__(self, factory):
        """

        :param factory: An instance of :class:`autobahn.twisted.websocket.WebSocketServerFactory`.
        :type factory: obj
        """
        self._factory = factory

    # noinspection PyUnusedLocal
    def getChildWithDefault(self, name, request):
        """
        This resource cannot have children, hence this will always fail.
        """
        return notFound(message="No such child resource.")

    def putChild(self, path, child):
        """
        This resource cannot have children, hence this is always ignored.
        """

    def render(self, request):
        """
        Render the resource. This will takeover the transport underlying
        the request, create a :class:`autobahn.twisted.websocket.WebSocketServerProtocol`
        and let that do any subsequent communication.
        """
        # for reasons unknown, the transport is already None when the
        # request is over HTTP2. request.channel.getPeer() is valid at
        # this point however
        if request.channel.transport is None:
            # render an "error, yo're doing HTTPS over WSS" webpage
            from autobahn.websocket import protocol

            request.setResponseCode(426, b"Upgrade required")
            # RFC says MUST set upgrade along with 426 code:
            # https://tools.ietf.org/html/rfc7231#section-6.5.15
            request.setHeader(b"Upgrade", b"WebSocket")
            html = protocol._SERVER_STATUS_TEMPLATE % ("", protocol.__version__)
            return html.encode("utf8")

        # Create Autobahn WebSocket protocol.
        #
        protocol = self._factory.buildProtocol(request.transport.getPeer())
        if not protocol:
            # If protocol creation fails, we signal "internal server error"
            request.setResponseCode(500)
            return b""

        # Take over the transport from Twisted Web
        #
        transport, request.channel.transport = request.channel.transport, None

        # Connect the transport to our protocol. Once #3204 is fixed, there
        # may be a cleaner way of doing this.
        # http://twistedmatrix.com/trac/ticket/3204
        #
        if isinstance(transport, ProtocolWrapper):
            # i.e. TLS is a wrapping protocol
            transport.wrappedProtocol = protocol
        elif isinstance(transport.protocol, ProtocolWrapper):
            # this happens in new-TLS
            transport.protocol.wrappedProtocol = protocol
        else:
            transport.protocol = protocol
        protocol.makeConnection(transport)

        # On Twisted 16+, the transport is paused whilst the existing
        # request is served; there won't be any requests after us so
        # we can just resume this ourselves.
        # 17.1 version
        if hasattr(transport, "_networkProducer"):
            transport._networkProducer.resumeProducing()
        # 16.x version
        elif hasattr(transport, "resumeProducing"):
            transport.resumeProducing()

        # We recreate the request and forward the raw data. This is somewhat
        # silly (since Twisted Web already did the HTTP request parsing
        # which we will do a 2nd time), but it's totally non-invasive to our
        # code. Maybe improve this.
        #
        data = request.method + b" " + request.uri + b" HTTP/1.1\x0d\x0a"
        for h in request.requestHeaders.getAllRawHeaders():
            data += h[0] + b": " + b",".join(h[1]) + b"\x0d\x0a"
        data += b"\x0d\x0a"
        data += request.content.read()
        protocol.dataReceived(data)

        return NOT_DONE_YET



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

# Note: This code is a Python implementation of the algorithm
# "Flexible and Economical UTF-8 Decoder" by Bjoern Hoehrmann
# bjoern@hoehrmann.de, http://bjoern.hoehrmann.de/utf-8/decoder/dfa/

__all__ = ("Utf8Validator",)


# DFA transitions
UTF8VALIDATOR_DFA = (
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,  # 00..1f
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,  # 20..3f
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,  # 40..5f
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,  # 60..7f
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    9,
    9,
    9,
    9,
    9,
    9,
    9,
    9,
    9,
    9,
    9,
    9,
    9,
    9,
    9,
    9,  # 80..9f
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,
    7,  # a0..bf
    8,
    8,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,  # c0..df
    0xA,
    0x3,
    0x3,
    0x3,
    0x3,
    0x3,
    0x3,
    0x3,
    0x3,
    0x3,
    0x3,
    0x3,
    0x3,
    0x4,
    0x3,
    0x3,  # e0..ef
    0xB,
    0x6,
    0x6,
    0x6,
    0x5,
    0x8,
    0x8,
    0x8,
    0x8,
    0x8,
    0x8,
    0x8,
    0x8,
    0x8,
    0x8,
    0x8,  # f0..ff
    0x0,
    0x1,
    0x2,
    0x3,
    0x5,
    0x8,
    0x7,
    0x1,
    0x1,
    0x1,
    0x4,
    0x6,
    0x1,
    0x1,
    0x1,
    0x1,  # s0..s0
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    0,
    1,
    1,
    1,
    1,
    1,
    0,
    1,
    0,
    1,
    1,
    1,
    1,
    1,
    1,  # s1..s2
    1,
    2,
    1,
    1,
    1,
    1,
    1,
    2,
    1,
    2,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    2,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,  # s3..s4
    1,
    2,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    2,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    3,
    1,
    3,
    1,
    1,
    1,
    1,
    1,
    1,  # s5..s6
    1,
    3,
    1,
    1,
    1,
    1,
    1,
    3,
    1,
    3,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    3,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,  # s7..s8
)

UTF8_ACCEPT = 0
UTF8_REJECT = 1


# Import USES_NVX flag from parent module
from autobahn.websocket import USES_NVX

if USES_NVX:
    # Use NVX native implementation (CFFI-based, works on CPython and PyPy)
    from autobahn.nvx._utf8validator import Utf8Validator
else:
    # Use pure Python fallback implementation
    #
    # Do NOT touch this code unless you know what you are doing!
    # https://github.com/oberstet/scratchbox/tree/master/python/utf8

    # Python 3 and above

    # convert DFA table to bytes (performance)
    UTF8VALIDATOR_DFA_S = bytes(UTF8VALIDATOR_DFA)

    class Utf8Validator(object):
        """
        Incremental UTF-8 validator with constant memory consumption (minimal state).

        Implements the algorithm "Flexible and Economical UTF-8 Decoder" by
        Bjoern Hoehrmann (http://bjoern.hoehrmann.de/utf-8/decoder/dfa/).
        """

        __slots__ = (
            "_codepoint",
            "_state",
            "_index",
        )

        def __init__(self):
            self._codepoint = None
            self._state = None
            self._index = None
            self.reset()

        def decode(self, b):
            """
            Eat one UTF-8 octet, and validate on the fly.

            Returns ``UTF8_ACCEPT`` when enough octets have been consumed, in which case
            ``self.codepoint`` contains the decoded Unicode code point.

            Returns ``UTF8_REJECT`` when invalid UTF-8 was encountered.

            Returns some other positive integer when more octets need to be eaten.
            """
            tt = UTF8VALIDATOR_DFA_S[b]
            if self._state != UTF8_ACCEPT:
                self._codepoint = (b & 0x3F) | (self._codepoint << 6)
            else:
                self._codepoint = (0xFF >> tt) & b
            self._state = UTF8VALIDATOR_DFA_S[256 + self._state * 16 + tt]
            return self._state

        def reset(self):
            """
            Reset validator to start new incremental UTF-8 decode/validation.
            """
            self._state = UTF8_ACCEPT  # the empty string is valid UTF8
            self._codepoint = 0
            self._index = 0

        def validate(self, ba):
            """
            Incrementally validate a chunk of bytes provided as string.

            Will return a quad ``(valid?, endsOnCodePoint?, currentIndex, totalIndex)``.

            As soon as an octet is encountered which renders the octet sequence
            invalid, a quad with ``valid? == False`` is returned. ``currentIndex`` returns
            the index within the currently consumed chunk, and ``totalIndex`` the
            index within the total consumed sequence that was the point of bail out.
            When ``valid? == True``, currentIndex will be ``len(ba)`` and ``totalIndex`` the
            total amount of consumed bytes.
            """
            #
            # The code here is written for optimal JITting in PyPy, not for best
            # readability by your grandma or particular elegance. Do NOT touch!
            #
            l = len(ba)
            i = 0
            state = self._state
            while i < l:
                # optimized version of decode(), since we are not interested in actual code points
                state = UTF8VALIDATOR_DFA_S[
                    256 + (state << 4) + UTF8VALIDATOR_DFA_S[ba[i]]
                ]
                if state == UTF8_REJECT:
                    self._state = state
                    self._index += i
                    return False, False, i, self._index
                i += 1
            self._state = state
            self._index += l
            return True, state == UTF8_ACCEPT, l, self._index



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import zlib

from autobahn.util import public
from autobahn.websocket.compress_base import (
    PerMessageCompress,
    PerMessageCompressOffer,
    PerMessageCompressOfferAccept,
    PerMessageCompressResponse,
    PerMessageCompressResponseAccept,
)

__all__ = (
    "PerMessageDeflate",
    "PerMessageDeflateMixin",
    "PerMessageDeflateOffer",
    "PerMessageDeflateOfferAccept",
    "PerMessageDeflateResponse",
    "PerMessageDeflateResponseAccept",
)


class PerMessageDeflateMixin(object):
    """
    Mixin class for this extension.
    """

    EXTENSION_NAME = "permessage-deflate"
    """
    Name of this WebSocket extension.
    """

    WINDOW_SIZE_PERMISSIBLE_VALUES = [9, 10, 11, 12, 13, 14, 15]
    """
    Permissible value for window size parameter.
    Higher values use more memory, but produce smaller output. The default is 15.
    """

    MEM_LEVEL_PERMISSIBLE_VALUES = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    """
    Permissible value for memory level parameter.
    Higher values use more memory, but are faster and produce smaller output. The default is 8.
    """


@public
class PerMessageDeflateOffer(PerMessageCompressOffer, PerMessageDeflateMixin):
    """
    Set of extension parameters for `permessage-deflate` WebSocket extension
    offered by a client to a server.
    """

    @classmethod
    def parse(cls, params):
        """
        Parses a WebSocket extension offer for `permessage-deflate` provided by a client to a server.

        :param params: Output from :func:`autobahn.websocket.WebSocketProtocol._parseExtensionsHeader`.
        :type params: list

        :returns: A new instance of :class:`autobahn.compress.PerMessageDeflateOffer`.
        :rtype: obj
        """

        # extension parameter defaults
        accept_max_window_bits = False
        accept_no_context_takeover = True
        # accept_no_context_takeover = False # FIXME: this may change in draft
        request_max_window_bits = 0
        request_no_context_takeover = False

        # verify/parse client ("client-to-server direction") parameters of permessage-deflate offer
        for p in params:
            if len(params[p]) > 1:
                raise Exception(
                    "multiple occurrence of extension parameter '%s' for extension '%s'"
                    % (p, cls.EXTENSION_NAME)
                )

            val = params[p][0]

            if p == "client_max_window_bits":
                #
                # see: https://tools.ietf.org/html/draft-ietf-hybi-permessage-compression-18
                # 8.1.2.2. client_max_window_bits

                # ".. This parameter has no value or a decimal integer value without
                # leading zeroes between 9 to 15 inclusive ..""

                # noinspection PySimplifyBooleanCheck
                if val is not True:
                    try:
                        val = int(val)
                    except:
                        raise Exception(
                            "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                            % (val, p, cls.EXTENSION_NAME)
                        )
                    else:
                        if (
                            val
                            not in PerMessageDeflateMixin.WINDOW_SIZE_PERMISSIBLE_VALUES
                        ):
                            raise Exception(
                                "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                                % (val, p, cls.EXTENSION_NAME)
                            )
                        else:
                            # FIXME (maybe): possibly forward/process the client hint!
                            # accept_max_window_bits = val
                            accept_max_window_bits = True
                else:
                    accept_max_window_bits = True

            elif p == "client_no_context_takeover":
                # noinspection PySimplifyBooleanCheck
                if val is not True:
                    raise Exception(
                        "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                        % (val, p, cls.EXTENSION_NAME)
                    )
                else:
                    accept_no_context_takeover = True

            elif p == "server_max_window_bits":
                try:
                    val = int(val)
                except:
                    raise Exception(
                        "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                        % (val, p, cls.EXTENSION_NAME)
                    )
                else:
                    if val not in PerMessageDeflateMixin.WINDOW_SIZE_PERMISSIBLE_VALUES:
                        raise Exception(
                            "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                            % (val, p, cls.EXTENSION_NAME)
                        )
                    else:
                        request_max_window_bits = val

            elif p == "server_no_context_takeover":
                # noinspection PySimplifyBooleanCheck
                if val is not True:
                    raise Exception(
                        "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                        % (val, p, cls.EXTENSION_NAME)
                    )
                else:
                    request_no_context_takeover = True

            else:
                raise Exception(
                    "illegal extension parameter '%s' for extension '%s'"
                    % (p, cls.EXTENSION_NAME)
                )

        offer = cls(
            accept_no_context_takeover,
            accept_max_window_bits,
            request_no_context_takeover,
            request_max_window_bits,
        )
        return offer

    def __init__(
        self,
        accept_no_context_takeover=True,
        accept_max_window_bits=True,
        request_no_context_takeover=False,
        request_max_window_bits=0,
    ):
        """

        :param accept_no_context_takeover: When ``True``, the client accepts the "no context takeover" feature.
        :type accept_no_context_takeover: bool
        :param accept_max_window_bits: When ``True``, the client accepts setting "max window size".
        :type accept_max_window_bits: bool
        :param request_no_context_takeover: When ``True``, the client request the "no context takeover" feature.
        :type request_no_context_takeover: bool
        :param request_max_window_bits: When non-zero, the client requests the given "max window size" (must be
            and integer from the interval ``[9..15]``).
        :type request_max_window_bits: int
        """
        if type(accept_no_context_takeover) != bool:
            raise Exception(
                "invalid type %s for accept_no_context_takeover"
                % type(accept_no_context_takeover)
            )

        self.accept_no_context_takeover = accept_no_context_takeover

        if type(accept_max_window_bits) != bool:
            raise Exception(
                "invalid type %s for accept_max_window_bits"
                % type(accept_max_window_bits)
            )

        self.accept_max_window_bits = accept_max_window_bits

        if type(request_no_context_takeover) != bool:
            raise Exception(
                "invalid type %s for request_no_context_takeover"
                % type(request_no_context_takeover)
            )

        self.request_no_context_takeover = request_no_context_takeover

        if (
            request_max_window_bits != 0
            and request_max_window_bits not in self.WINDOW_SIZE_PERMISSIBLE_VALUES
        ):
            raise Exception(
                "invalid value %s for request_max_window_bits - permissible values %s"
                % (request_max_window_bits, self.WINDOW_SIZE_PERMISSIBLE_VALUES)
            )

        self.request_max_window_bits = request_max_window_bits

    def get_extension_string(self):
        """
        Returns the WebSocket extension configuration string as sent to the server.

        :returns: PMCE configuration string.
        :rtype: str
        """
        pmce_string = self.EXTENSION_NAME
        if self.accept_no_context_takeover:
            pmce_string += "; client_no_context_takeover"
        if self.accept_max_window_bits:
            pmce_string += "; client_max_window_bits"
        if self.request_no_context_takeover:
            pmce_string += "; server_no_context_takeover"
        if self.request_max_window_bits != 0:
            pmce_string += "; server_max_window_bits=%d" % self.request_max_window_bits
        return pmce_string

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: JSON serializable representation.
        :rtype: dict
        """
        return {
            "extension": self.EXTENSION_NAME,
            "accept_no_context_takeover": self.accept_no_context_takeover,
            "accept_max_window_bits": self.accept_max_window_bits,
            "request_no_context_takeover": self.request_no_context_takeover,
            "request_max_window_bits": self.request_max_window_bits,
        }

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: Python string representation.
        :rtype: str
        """
        return (
            "PerMessageDeflateOffer(accept_no_context_takeover = %s, accept_max_window_bits = %s, request_no_context_takeover = %s, request_max_window_bits = %s)"
            % (
                self.accept_no_context_takeover,
                self.accept_max_window_bits,
                self.request_no_context_takeover,
                self.request_max_window_bits,
            )
        )


@public
class PerMessageDeflateOfferAccept(
    PerMessageCompressOfferAccept, PerMessageDeflateMixin
):
    """
    Set of parameters with which to accept an `permessage-deflate` offer
    from a client by a server.
    """

    def __init__(
        self,
        offer,
        request_no_context_takeover=False,
        request_max_window_bits=0,
        no_context_takeover=None,
        window_bits=None,
        mem_level=None,
        max_message_size=None,
    ):
        """

        :param offer: The offer being accepted.
        :type offer: Instance of :class:`autobahn.compress.PerMessageDeflateOffer`.
        :param request_no_context_takeover: When ``True``, the server requests the "no context takeover" feature.
        :type request_no_context_takeover: bool
        :param request_max_window_bits: When non-zero, the server requests the given "max window size" (must be
            and integer from the interval ``[9..15]``).
        :param request_max_window_bits: int
        :param no_context_takeover: Override server ("server-to-client direction") context takeover (this must
                be compatible with the offer).
        :type no_context_takeover: bool
        :param window_bits: Override server ("server-to-client direction") window size (this must be
                compatible with the offer).
        :type window_bits: int
        :param mem_level: Set server ("server-to-client direction") memory level.
        :type mem_level: int
        """
        if not isinstance(offer, PerMessageDeflateOffer):
            raise Exception("invalid type %s for offer" % type(offer))

        self.offer = offer

        if type(request_no_context_takeover) != bool:
            raise Exception(
                "invalid type %s for request_no_context_takeover"
                % type(request_no_context_takeover)
            )

        if request_no_context_takeover and not offer.accept_no_context_takeover:
            raise Exception(
                "invalid value %s for request_no_context_takeover - feature unsupported by client"
                % request_no_context_takeover
            )

        self.request_no_context_takeover = request_no_context_takeover

        if (
            request_max_window_bits != 0
            and request_max_window_bits not in self.WINDOW_SIZE_PERMISSIBLE_VALUES
        ):
            raise Exception(
                "invalid value %s for request_max_window_bits - permissible values %s"
                % (request_max_window_bits, self.WINDOW_SIZE_PERMISSIBLE_VALUES)
            )

        if request_max_window_bits != 0 and not offer.accept_max_window_bits:
            raise Exception(
                "invalid value %s for request_max_window_bits - feature unsupported by client"
                % request_max_window_bits
            )

        self.request_max_window_bits = request_max_window_bits

        if no_context_takeover is not None:
            if type(no_context_takeover) != bool:
                raise Exception(
                    "invalid type %s for no_context_takeover"
                    % type(no_context_takeover)
                )

            if offer.request_no_context_takeover and not no_context_takeover:
                raise Exception(
                    "invalid value %s for no_context_takeover - client requested feature"
                    % no_context_takeover
                )

        self.no_context_takeover = no_context_takeover

        if window_bits is not None:
            if window_bits not in self.WINDOW_SIZE_PERMISSIBLE_VALUES:
                raise Exception(
                    "invalid value %s for window_bits - permissible values %s"
                    % (window_bits, self.WINDOW_SIZE_PERMISSIBLE_VALUES)
                )

            if (
                offer.request_max_window_bits != 0
                and window_bits > offer.request_max_window_bits
            ):
                raise Exception(
                    "invalid value %s for window_bits - client requested lower maximum value"
                    % window_bits
                )

        self.window_bits = window_bits

        if mem_level is not None:
            if mem_level not in self.MEM_LEVEL_PERMISSIBLE_VALUES:
                raise Exception(
                    "invalid value %s for mem_level - permissible values %s"
                    % (mem_level, self.MEM_LEVEL_PERMISSIBLE_VALUES)
                )

        self.mem_level = mem_level
        self.max_message_size = max_message_size  # clamp/check values..?

    def get_extension_string(self):
        """
        Returns the WebSocket extension configuration string as sent to the server.

        :returns: PMCE configuration string.
        :rtype: str
        """
        pmce_string = self.EXTENSION_NAME
        if self.offer.request_no_context_takeover:
            pmce_string += "; server_no_context_takeover"
        if self.offer.request_max_window_bits != 0:
            pmce_string += (
                "; server_max_window_bits=%d" % self.offer.request_max_window_bits
            )
        if self.request_no_context_takeover:
            pmce_string += "; client_no_context_takeover"
        if self.request_max_window_bits != 0:
            pmce_string += "; client_max_window_bits=%d" % self.request_max_window_bits
        return pmce_string

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: JSON serializable representation.
        :rtype: dict
        """
        return {
            "extension": self.EXTENSION_NAME,
            "offer": self.offer.__json__(),
            "request_no_context_takeover": self.request_no_context_takeover,
            "request_max_window_bits": self.request_max_window_bits,
            "no_context_takeover": self.no_context_takeover,
            "window_bits": self.window_bits,
            "mem_level": self.mem_level,
            "max_message_size": self.max_message_size,
        }

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: Python string representation.
        :rtype: str
        """
        return (
            "PerMessageDeflateOfferAccept(offer = %s, request_no_context_takeover = %s, request_max_window_bits = %s, no_context_takeover = %s, window_bits = %s, mem_level = %s, max_message_size = %s)"
            % (
                self.offer.__repr__(),
                self.request_no_context_takeover,
                self.request_max_window_bits,
                self.no_context_takeover,
                self.window_bits,
                self.mem_level,
                self.max_message_size,
            )
        )


@public
class PerMessageDeflateResponse(PerMessageCompressResponse, PerMessageDeflateMixin):
    """
    Set of parameters for `permessage-deflate` responded by server.
    """

    @classmethod
    def parse(cls, params):
        """
        Parses a WebSocket extension response for `permessage-deflate` provided by a server to a client.

        :param params: Output from :func:`autobahn.websocket.WebSocketProtocol._parseExtensionsHeader`.
        :type params: list

        :returns: A new instance of :class:`autobahn.compress.PerMessageDeflateResponse`.
        :rtype: obj
        """
        client_max_window_bits = 0
        client_no_context_takeover = False
        server_max_window_bits = 0
        server_no_context_takeover = False

        for p in params:
            if len(params[p]) > 1:
                raise Exception(
                    "multiple occurrence of extension parameter '%s' for extension '%s'"
                    % (p, cls.EXTENSION_NAME)
                )

            val = params[p][0]

            if p == "client_max_window_bits":
                try:
                    val = int(val)
                except:
                    raise Exception(
                        "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                        % (val, p, cls.EXTENSION_NAME)
                    )
                else:
                    if val not in PerMessageDeflateMixin.WINDOW_SIZE_PERMISSIBLE_VALUES:
                        raise Exception(
                            "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                            % (val, p, cls.EXTENSION_NAME)
                        )
                    else:
                        client_max_window_bits = val

            elif p == "client_no_context_takeover":
                # noinspection PySimplifyBooleanCheck
                if val is not True:
                    raise Exception(
                        "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                        % (val, p, cls.EXTENSION_NAME)
                    )
                else:
                    client_no_context_takeover = True

            elif p == "server_max_window_bits":
                try:
                    val = int(val)
                except:
                    raise Exception(
                        "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                        % (val, p, cls.EXTENSION_NAME)
                    )
                else:
                    if val not in PerMessageDeflateMixin.WINDOW_SIZE_PERMISSIBLE_VALUES:
                        raise Exception(
                            "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                            % (val, p, cls.EXTENSION_NAME)
                        )
                    else:
                        server_max_window_bits = val

            elif p == "server_no_context_takeover":
                # noinspection PySimplifyBooleanCheck
                if val is not True:
                    raise Exception(
                        "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                        % (val, p, cls.EXTENSION_NAME)
                    )
                else:
                    server_no_context_takeover = True

            else:
                raise Exception(
                    "illegal extension parameter '%s' for extension '%s'"
                    % (p, cls.EXTENSION_NAME)
                )

        response = cls(
            client_max_window_bits,
            client_no_context_takeover,
            server_max_window_bits,
            server_no_context_takeover,
        )
        return response

    def __init__(
        self,
        client_max_window_bits,
        client_no_context_takeover,
        server_max_window_bits,
        server_no_context_takeover,
    ):
        """

        :param client_max_window_bits: FIXME
        :type client_max_window_bits: int
        :param client_no_context_takeover: FIXME
        :type client_no_context_takeover: bool
        :param server_max_window_bits: FIXME
        :type server_max_window_bits: int
        :param server_no_context_takeover: FIXME
        :type server_no_context_takeover: bool
        """
        self.client_max_window_bits = client_max_window_bits
        self.client_no_context_takeover = client_no_context_takeover
        self.server_max_window_bits = server_max_window_bits
        self.server_no_context_takeover = server_no_context_takeover

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: JSON serializable representation.
        :rtype: dict
        """
        return {
            "extension": self.EXTENSION_NAME,
            "client_max_window_bits": self.client_max_window_bits,
            "client_no_context_takeover": self.client_no_context_takeover,
            "server_max_window_bits": self.server_max_window_bits,
            "server_no_context_takeover": self.server_no_context_takeover,
        }

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: Python string representation.
        :rtype: str
        """
        return (
            "PerMessageDeflateResponse(client_max_window_bits = %s, client_no_context_takeover = %s, server_max_window_bits = %s, server_no_context_takeover = %s)"
            % (
                self.client_max_window_bits,
                self.client_no_context_takeover,
                self.server_max_window_bits,
                self.server_no_context_takeover,
            )
        )


@public
class PerMessageDeflateResponseAccept(
    PerMessageCompressResponseAccept, PerMessageDeflateMixin
):
    """
    Set of parameters with which to accept an `permessage-deflate` response
    from a server by a client.
    """

    def __init__(
        self,
        response,
        no_context_takeover=None,
        window_bits=None,
        mem_level=None,
        max_message_size=None,
    ):
        """

        :param response: The response being accepted.
        :type response: Instance of :class:`autobahn.compress.PerMessageDeflateResponse`.
        :param no_context_takeover: Override client ("client-to-server direction") context takeover (this must be compatible with response).
        :type no_context_takeover: bool
        :param window_bits: Override client ("client-to-server direction") window size (this must be compatible with response).
        :type window_bits: int
        :param mem_level: Set client ("client-to-server direction") memory level.
        :type mem_level: int
        """
        if not isinstance(response, PerMessageDeflateResponse):
            raise Exception("invalid type %s for response" % type(response))

        self.response = response

        if no_context_takeover is not None:
            if type(no_context_takeover) != bool:
                raise Exception(
                    "invalid type %s for no_context_takeover"
                    % type(no_context_takeover)
                )

            if response.client_no_context_takeover and not no_context_takeover:
                raise Exception(
                    "invalid value %s for no_context_takeover - server requested feature"
                    % no_context_takeover
                )

        self.no_context_takeover = no_context_takeover

        if window_bits is not None:
            if window_bits not in self.WINDOW_SIZE_PERMISSIBLE_VALUES:
                raise Exception(
                    "invalid value %s for window_bits - permissible values %s"
                    % (window_bits, self.WINDOW_SIZE_PERMISSIBLE_VALUES)
                )

            if (
                response.client_max_window_bits != 0
                and window_bits > response.client_max_window_bits
            ):
                raise Exception(
                    "invalid value %s for window_bits - server requested lower maximum value"
                    % window_bits
                )

        self.window_bits = window_bits

        if mem_level is not None:
            if mem_level not in self.MEM_LEVEL_PERMISSIBLE_VALUES:
                raise Exception(
                    "invalid value %s for mem_level - permissible values %s"
                    % (mem_level, self.MEM_LEVEL_PERMISSIBLE_VALUES)
                )

        self.mem_level = mem_level
        self.max_message_size = max_message_size

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: JSON serializable representation.
        :rtype: dict
        """
        return {
            "extension": self.EXTENSION_NAME,
            "response": self.response.__json__(),
            "no_context_takeover": self.no_context_takeover,
            "window_bits": self.window_bits,
            "mem_level": self.mem_level,
        }

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: Python string representation.
        :rtype: str
        """
        return (
            "PerMessageDeflateResponseAccept(response = %s, no_context_takeover = %s, window_bits = %s, mem_level = %s)"
            % (
                self.response.__repr__(),
                self.no_context_takeover,
                self.window_bits,
                self.mem_level,
            )
        )


# noinspection PyArgumentList
class PerMessageDeflate(PerMessageCompress, PerMessageDeflateMixin):
    """
    `permessage-deflate` WebSocket extension processor.
    """

    DEFAULT_WINDOW_BITS = zlib.MAX_WBITS
    DEFAULT_MEM_LEVEL = 8

    @classmethod
    def create_from_response_accept(cls, is_server, accept):
        # accept: instance of PerMessageDeflateResponseAccept
        pmce = cls(
            is_server,
            accept.response.server_no_context_takeover,
            (
                accept.no_context_takeover
                if accept.no_context_takeover is not None
                else accept.response.client_no_context_takeover
            ),
            accept.response.server_max_window_bits,
            (
                accept.window_bits
                if accept.window_bits is not None
                else accept.response.client_max_window_bits
            ),
            accept.mem_level,
            accept.max_message_size,
        )
        return pmce

    @classmethod
    def create_from_offer_accept(cls, is_server, accept):
        # accept: instance of PerMessageDeflateOfferAccept
        pmce = cls(
            is_server,
            (
                accept.no_context_takeover
                if accept.no_context_takeover is not None
                else accept.offer.request_no_context_takeover
            ),
            accept.request_no_context_takeover,
            (
                accept.window_bits
                if accept.window_bits is not None
                else accept.offer.request_max_window_bits
            ),
            accept.request_max_window_bits,
            accept.mem_level,
            accept.max_message_size,
        )
        return pmce

    def __init__(
        self,
        is_server,
        server_no_context_takeover,
        client_no_context_takeover,
        server_max_window_bits,
        client_max_window_bits,
        mem_level,
        max_message_size=None,
    ):
        self._is_server = is_server

        self.server_no_context_takeover = server_no_context_takeover
        self.client_no_context_takeover = client_no_context_takeover

        self.server_max_window_bits = (
            server_max_window_bits
            if server_max_window_bits != 0
            else self.DEFAULT_WINDOW_BITS
        )
        self.client_max_window_bits = (
            client_max_window_bits
            if client_max_window_bits != 0
            else self.DEFAULT_WINDOW_BITS
        )

        self.mem_level = mem_level if mem_level else self.DEFAULT_MEM_LEVEL
        self.max_message_size = max_message_size  # None means "no limit"

        self._compressor = None
        self._decompressor = None

    def __json__(self):
        return {
            "extension": self.EXTENSION_NAME,
            "is_server": self._is_server,
            "server_no_context_takeover": self.server_no_context_takeover,
            "client_no_context_takeover": self.client_no_context_takeover,
            "server_max_window_bits": self.server_max_window_bits,
            "client_max_window_bits": self.client_max_window_bits,
            "mem_level": self.mem_level,
        }

    def __repr__(self):
        return (
            "PerMessageDeflate(is_server = %s, server_no_context_takeover = %s, client_no_context_takeover = %s, server_max_window_bits = %s, client_max_window_bits = %s, mem_level = %s)"
            % (
                self._is_server,
                self.server_no_context_takeover,
                self.client_no_context_takeover,
                self.server_max_window_bits,
                self.client_max_window_bits,
                self.mem_level,
            )
        )

    def start_compress_message(self):
        # compressobj([level[, method[, wbits[, mem_level[, strategy]]]]])
        # http://bugs.python.org/issue19278
        # http://hg.python.org/cpython/rev/c54c8e71b79a
        if self._is_server:
            if self._compressor is None or self.server_no_context_takeover:
                self._compressor = zlib.compressobj(
                    zlib.Z_DEFAULT_COMPRESSION,
                    zlib.DEFLATED,
                    -self.server_max_window_bits,
                    self.mem_level,
                )
        else:
            if self._compressor is None or self.client_no_context_takeover:
                self._compressor = zlib.compressobj(
                    zlib.Z_DEFAULT_COMPRESSION,
                    zlib.DEFLATED,
                    -self.client_max_window_bits,
                    self.mem_level,
                )

    def compress_message_data(self, data):
        return self._compressor.compress(data)

    def end_compress_message(self):
        data = self._compressor.flush(zlib.Z_SYNC_FLUSH)
        return data[:-4]

    def start_decompress_message(self):
        if self._is_server:
            if self._decompressor is None or self.client_no_context_takeover:
                self._decompressor = zlib.decompressobj(-self.client_max_window_bits)
        else:
            if self._decompressor is None or self.server_no_context_takeover:
                self._decompressor = zlib.decompressobj(-self.server_max_window_bits)

    def decompress_message_data(self, data):
        if self.max_message_size is not None:
            return self._decompressor.decompress(data, self.max_message_size)
        return self._decompressor.decompress(data)

    def end_decompress_message(self):
        # Eat stripped LEN and NLEN field of a non-compressed block added
        # for Z_SYNC_FLUSH.
        self._decompressor.decompress(b"\x00\x00\xff\xff")



import os
from unittest.mock import Mock, call

import pytest

from autobahn.asyncio.rawsocket import (
    PrefixProtocol,
    RawSocketClientProtocol,
    RawSocketServerProtocol,
    WampRawSocketClientFactory,
    WampRawSocketServerFactory,
)
from autobahn.asyncio.util import get_serializers
from autobahn.wamp import message
from autobahn.wamp.types import TransportDetails


@pytest.mark.skipif(
    not os.environ.get("USE_ASYNCIO", False), reason="test runs on asyncio only"
)
def test_sers():
    serializers = get_serializers()
    assert len(serializers) > 0
    m = serializers[0]().serialize(message.Abort("close"))
    assert m


@pytest.mark.skipif(
    not os.environ.get("USE_ASYNCIO", False), reason="test runs on asyncio only"
)
def test_prefix():
    p = PrefixProtocol()
    transport = Mock()
    receiver = Mock()
    p.stringReceived = receiver
    p.connection_made(transport)
    small_msg = b"\x00\x00\x00\x04abcd"
    p.data_received(small_msg)
    receiver.assert_called_once_with(b"abcd")
    assert len(p._buffer) == 0

    p.sendString(b"abcd")

    # print(transport.write.call_args_list)
    transport.write.assert_has_calls([call(b"\x00\x00\x00\x04"), call(b"abcd")])

    transport.reset_mock()
    receiver.reset_mock()

    big_msg = b"\x00\x00\x00\x0c" + b"0123456789AB"

    p.data_received(big_msg[0:2])
    assert not receiver.called

    p.data_received(big_msg[2:6])
    assert not receiver.called

    p.data_received(big_msg[6:11])
    assert not receiver.called

    p.data_received(big_msg[11:16])
    receiver.assert_called_once_with(b"0123456789AB")

    transport.reset_mock()
    receiver.reset_mock()

    two_messages = (
        b"\x00\x00\x00\x04" + b"abcd" + b"\x00\x00\x00\x05" + b"12345" + b"\x00"
    )
    p.data_received(two_messages)
    receiver.assert_has_calls([call(b"abcd"), call(b"12345")])
    assert p._buffer == b"\x00"


@pytest.mark.skipif(
    not os.environ.get("USE_ASYNCIO", False), reason="test runs on asyncio only"
)
def test_is_closed():
    class CP(RawSocketClientProtocol):
        @property
        def serializer_id(self):
            return 1

    client = CP()

    on_hs = Mock()
    transport = Mock()
    receiver = Mock()
    client.stringReceived = receiver
    client._on_handshake_complete = on_hs
    assert client.is_closed.done()
    client.connection_made(transport)
    assert not client.is_closed.done()
    client.connection_lost(None)

    assert client.is_closed.done()


@pytest.mark.skipif(
    not os.environ.get("USE_ASYNCIO", False), reason="test runs on asyncio only"
)
def test_raw_socket_server1():
    server = RawSocketServerProtocol()
    ser = Mock(return_value=True)
    on_hs = Mock()
    transport = Mock()
    receiver = Mock()
    server.supports_serializer = ser
    server.stringReceived = receiver
    server._on_handshake_complete = on_hs
    server.stringReceived = receiver

    server.connection_made(transport)
    hs = b"\x7f\xf1\x00\x00" + b"\x00\x00\x00\x04abcd"
    server.data_received(hs)

    ser.assert_called_once_with(1)
    on_hs.assert_called_once_with()
    assert transport.write.called
    transport.write.assert_called_once_with(b"\x7f\xf1\x00\x00")
    assert not transport.close.called
    receiver.assert_called_once_with(b"abcd")


@pytest.mark.skipif(
    not os.environ.get("USE_ASYNCIO", False), reason="test runs on asyncio only"
)
def test_raw_socket_server_errors():
    server = RawSocketServerProtocol()
    ser = Mock(return_value=True)
    on_hs = Mock()
    transport = Mock()
    receiver = Mock()
    server.supports_serializer = ser
    server.stringReceived = receiver
    server._on_handshake_complete = on_hs
    server.stringReceived = receiver
    server.connection_made(transport)
    server.data_received(b"abcdef")
    transport.close.assert_called_once_with()

    server = RawSocketServerProtocol()
    ser = Mock(return_value=False)
    on_hs = Mock()
    transport = Mock(spec_set=("close", "write", "get_extra_info"))
    receiver = Mock()
    server.supports_serializer = ser
    server.stringReceived = receiver
    server._on_handshake_complete = on_hs
    server.stringReceived = receiver
    server.connection_made(transport)
    server.data_received(b"\x7f\xf1\x00\x00")
    transport.close.assert_called_once_with()
    transport.write.assert_called_once_with(b"\x7f\x10\x00\x00")


@pytest.mark.skipif(
    not os.environ.get("USE_ASYNCIO", False), reason="test runs on asyncio only"
)
def test_raw_socket_client1():
    class CP(RawSocketClientProtocol):
        @property
        def serializer_id(self):
            return 1

    client = CP()

    on_hs = Mock()
    transport = Mock()
    receiver = Mock()
    client.stringReceived = receiver
    client._on_handshake_complete = on_hs

    client.connection_made(transport)
    client.data_received(b"\x7f\xf1\x00\x00" + b"\x00\x00\x00\x04abcd")
    on_hs.assert_called_once_with()
    assert transport.write.called
    transport.write.called_one_with(b"\x7f\xf1\x00\x00")
    assert not transport.close.called
    receiver.assert_called_once_with(b"abcd")


@pytest.mark.skipif(
    not os.environ.get("USE_ASYNCIO", False), reason="test runs on asyncio only"
)
def test_raw_socket_client_error():
    class CP(RawSocketClientProtocol):
        @property
        def serializer_id(self):
            return 1

    client = CP()

    on_hs = Mock()
    transport = Mock(spec_set=("close", "write", "get_extra_info"))
    receiver = Mock()
    client.stringReceived = receiver
    client._on_handshake_complete = on_hs
    client.connection_made(transport)

    client.data_received(b"\x7f\xf1\x00\x01")
    transport.close.assert_called_once_with()


@pytest.mark.skipif(
    not os.environ.get("USE_ASYNCIO", False), reason="test runs on asyncio only"
)
def test_wamp_server():
    transport = Mock(spec_set=("abort", "close", "write", "get_extra_info"))
    transport.write = Mock(side_effect=lambda m: messages.append(m))
    server = Mock(spec=["onOpen", "onMessage"])

    def fact_server():
        return server

    messages = []

    proto = WampRawSocketServerFactory(fact_server)()
    proto.connection_made(transport)
    assert proto.transport_details.is_server is True
    assert (
        proto.transport_details.channel_framing
        == TransportDetails.CHANNEL_FRAMING_RAWSOCKET
    )
    assert proto.factory._serializers
    s = proto.factory._serializers[1].RAWSOCKET_SERIALIZER_ID
    proto.data_received(bytes(bytearray([0x7F, 0xF0 | s, 0, 0])))
    assert proto._serializer
    server.onOpen.assert_called_once_with(proto)

    proto.send(message.Abort("close"))
    for d in messages[1:]:
        proto.data_received(d)
    assert server.onMessage.called
    assert isinstance(server.onMessage.call_args[0][0], message.Abort)


@pytest.mark.skipif(
    not os.environ.get("USE_ASYNCIO", False), reason="test runs on asyncio only"
)
def test_wamp_client():
    transport = Mock(spec_set=("abort", "close", "write", "get_extra_info"))
    transport.write = Mock(side_effect=lambda m: messages.append(m))
    client = Mock(spec=["onOpen", "onMessage"])

    def fact_client():
        return client

    messages = []

    proto = WampRawSocketClientFactory(fact_client)()
    proto.connection_made(transport)
    assert proto.transport_details.is_server is False
    assert (
        proto.transport_details.channel_framing
        == TransportDetails.CHANNEL_FRAMING_RAWSOCKET
    )
    assert proto._serializer
    s = proto._serializer.RAWSOCKET_SERIALIZER_ID
    proto.data_received(bytes(bytearray([0x7F, 0xF0 | s, 0, 0])))
    client.onOpen.assert_called_once_with(proto)

    proto.send(message.Abort("close"))
    for d in messages[1:]:
        proto.data_received(d)
    assert client.onMessage.called
    assert isinstance(client.onMessage.call_args[0][0], message.Abort)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from twisted.trial.unittest import TestCase


class PluginTests(TestCase):
    if True:
        skip = "Plugins don't work under Python3 yet"

    def test_import(self):
        from twisted.plugins import autobahn_endpoints

        self.assertTrue(hasattr(autobahn_endpoints, "AutobahnClientParser"))

    def test_parse_client_basic(self):
        from twisted.plugins import autobahn_endpoints

        self.assertTrue(hasattr(autobahn_endpoints, "AutobahnClientParser"))
        from twisted.internet import reactor
        from twisted.internet.endpoints import clientFromString, quoteStringArgument

        ep_string = "autobahn:{0}:url={1}".format(
            quoteStringArgument("tcp:localhost:9000"),
            quoteStringArgument("ws://localhost:9000"),
        )
        # we're just testing that this doesn't fail entirely
        clientFromString(reactor, ep_string)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class CallerFeatures(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = CallerFeatures()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsCallerFeatures(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # CallerFeatures
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # CallerFeatures
    def CallerIdentification(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # CallerFeatures
    def CallTimeout(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # CallerFeatures
    def CallCanceling(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # CallerFeatures
    def ProgressiveCallResults(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # CallerFeatures
    def PayloadTransparency(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # CallerFeatures
    def PayloadEncryptionCryptobox(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False


def CallerFeaturesStart(builder):
    builder.StartObject(6)


def Start(builder):
    CallerFeaturesStart(builder)


def CallerFeaturesAddCallerIdentification(builder, callerIdentification):
    builder.PrependBoolSlot(0, callerIdentification, 0)


def AddCallerIdentification(builder, callerIdentification):
    CallerFeaturesAddCallerIdentification(builder, callerIdentification)


def CallerFeaturesAddCallTimeout(builder, callTimeout):
    builder.PrependBoolSlot(1, callTimeout, 0)


def AddCallTimeout(builder, callTimeout):
    CallerFeaturesAddCallTimeout(builder, callTimeout)


def CallerFeaturesAddCallCanceling(builder, callCanceling):
    builder.PrependBoolSlot(2, callCanceling, 0)


def AddCallCanceling(builder, callCanceling):
    CallerFeaturesAddCallCanceling(builder, callCanceling)


def CallerFeaturesAddProgressiveCallResults(builder, progressiveCallResults):
    builder.PrependBoolSlot(3, progressiveCallResults, 0)


def AddProgressiveCallResults(builder, progressiveCallResults):
    CallerFeaturesAddProgressiveCallResults(builder, progressiveCallResults)


def CallerFeaturesAddPayloadTransparency(builder, payloadTransparency):
    builder.PrependBoolSlot(4, payloadTransparency, 0)


def AddPayloadTransparency(builder, payloadTransparency):
    CallerFeaturesAddPayloadTransparency(builder, payloadTransparency)


def CallerFeaturesAddPayloadEncryptionCryptobox(builder, payloadEncryptionCryptobox):
    builder.PrependBoolSlot(5, payloadEncryptionCryptobox, 0)


def AddPayloadEncryptionCryptobox(builder, payloadEncryptionCryptobox):
    CallerFeaturesAddPayloadEncryptionCryptobox(builder, payloadEncryptionCryptobox)


def CallerFeaturesEnd(builder):
    return builder.EndObject()


def End(builder):
    return CallerFeaturesEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import binascii
import os
import struct
from binascii import a2b_hex, b2a_hex
from typing import Any, Callable, Dict, Optional, Union

import txaio

from autobahn import util
from autobahn.util import parse_keyfile
from autobahn.wamp.interfaces import ICryptosignKey, ISecurityModule
from autobahn.wamp.message import _URI_PAT_REALM_NAME_ETH
from autobahn.wamp.mnemonic import mnemonic_to_private_key
from autobahn.wamp.types import Challenge

__all__ = [
    "HAS_CRYPTOSIGN",
]

try:
    # try to import everything we need for WAMP-cryptosign
    from nacl import bindings, encoding, signing
    from nacl.signing import SignedMessage
except ImportError:
    HAS_CRYPTOSIGN = False
else:
    HAS_CRYPTOSIGN = True
    __all__.append("CryptosignKey")


def _unpack(keydata):
    """
    Unpack a SSH agent key blob into parts.

    See: http://blog.oddbit.com/2011/05/08/converting-openssh-public-keys/
    """
    parts = []
    while keydata:
        # read the length of the data
        dlen = struct.unpack(">I", keydata[:4])[0]

        # read in <length> bytes
        data, keydata = keydata[4 : dlen + 4], keydata[4 + dlen :]
        parts.append(data)
    return parts


def _pack(keyparts):
    """
    Pack parts into a SSH key blob.
    """
    parts = []
    for part in keyparts:
        parts.append(struct.pack(">I", len(part)))
        parts.append(part)
    return b"".join(parts)


def _read_ssh_ed25519_pubkey(keydata):
    """
    Parse an OpenSSH Ed25519 public key from a string into a raw public key.

    Example input:

        ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIJukDU5fqXv/yVhSirsDWsUFyOodZyCSLxyitPPzWJW9 oberstet@office-corei7

    :param keydata: The OpenSSH Ed25519 public key data to parse.
    :type keydata: str

    :returns: pair of raw public key (32 bytes) and comment
    :rtype: tuple
    """
    if type(keydata) != str:
        raise Exception("invalid type {} for keydata".format(type(keydata)))

    parts = keydata.strip().split()
    if len(parts) != 3:
        raise Exception("invalid SSH Ed25519 public key")
    algo, keydata, comment = parts

    if algo != "ssh-ed25519":
        raise Exception("not a Ed25519 SSH public key (but {})".format(algo))

    blob = binascii.a2b_base64(keydata)

    try:
        key = _unpack(blob)[1]
    except Exception as e:
        raise Exception("could not parse key ({})".format(e))

    if len(key) != 32:
        raise Exception(
            "invalid length {} for embedded raw key (must be 32 bytes)".format(len(key))
        )

    return key, comment


class _SSHPacketReader:
    """
    Read OpenSSH packet format which is used for key material.
    """

    def __init__(self, packet):
        self._packet = packet
        self._idx = 0
        self._len = len(packet)

    def get_remaining_payload(self):
        return self._packet[self._idx :]

    def get_bytes(self, size):
        if self._idx + size > self._len:
            raise Exception("incomplete packet")

        value = self._packet[self._idx : self._idx + size]
        self._idx += size
        return value

    def get_uint32(self):
        return struct.unpack(">I", self.get_bytes(4))[0]

    def get_string(self):
        return self.get_bytes(self.get_uint32())


def _makepad(size: int) -> bytes:
    assert 0 <= size < 255
    return b"".join(x.to_bytes(1, byteorder="big") for x in range(1, size + 1))


def _read_ssh_ed25519_privkey(keydata):
    """
    Parse an OpenSSH Ed25519 private key from a string into a raw private key.

    Example input:

        -----BEGIN OPENSSH PRIVATE KEY-----
        b3BlbnNzaC1rZXktdjEAAAAABG5vbmUAAAAEbm9uZQAAAAAAAAABAAAAMwAAAAtzc2gtZW
        QyNTUxOQAAACCbpA1OX6l7/8lYUoq7A1rFBcjqHWcgki8corTz81iVvQAAAKDWjZ0Y1o2d
        GAAAAAtzc2gtZWQyNTUxOQAAACCbpA1OX6l7/8lYUoq7A1rFBcjqHWcgki8corTz81iVvQ
        AAAEArodzIMjH9MOBz0X+HDvL06rEJOMYFhzGQ5zXPM7b7fZukDU5fqXv/yVhSirsDWsUF
        yOodZyCSLxyitPPzWJW9AAAAFm9iZXJzdGV0QG9mZmljZS1jb3JlaTcBAgMEBQYH
        -----END OPENSSH PRIVATE KEY-----


    :param keydata: The OpenSSH Ed25519 private key data to parse.
    :type keydata: str

    :returns: pair of raw private key (32 bytes) and comment
    :rtype: tuple
    """

    # Some pointers:
    # https://github.com/ronf/asyncssh/blob/master/asyncssh/public_key.py
    # https://github.com/ronf/asyncssh/blob/master/asyncssh/ed25519.py
    # crypto_sign_ed25519_sk_to_seed
    # https://github.com/jedisct1/libsodium/blob/master/src/libsodium/crypto_sign/ed25519/sign_ed25519_api.c#L27
    # https://tools.ietf.org/html/draft-bjh21-ssh-ed25519-02
    # http://blog.oddbit.com/2011/05/08/converting-openssh-public-keys/

    SSH_BEGIN = "-----BEGIN OPENSSH PRIVATE KEY-----"
    SSH_END = "-----END OPENSSH PRIVATE KEY-----"
    OPENSSH_KEY_V1 = b"openssh-key-v1\0"

    if not (keydata.startswith(SSH_BEGIN) and keydata.endswith(SSH_END)):
        raise Exception(
            "invalid OpenSSH private key (does not start/end with OPENSSH preamble)"
        )

    ssh_end = keydata.find(SSH_END)
    keydata = keydata[len(SSH_BEGIN) : ssh_end]
    keydata = "".join(x.strip() for x in keydata.split())
    blob = binascii.a2b_base64(keydata)

    blob = blob[len(OPENSSH_KEY_V1) :]
    packet = _SSHPacketReader(blob)

    cipher_name = packet.get_string()
    kdf = packet.get_string()
    packet.get_string()  # kdf_data
    nkeys = packet.get_uint32()
    packet.get_string()  # public_key
    key_data = packet.get_string()
    mac = packet.get_remaining_payload()

    block_size = 8

    if cipher_name != b"none":
        raise Exception(
            "encrypted private keys not supported (please remove the passphrase from your private key or use SSH agent)"
        )

    if kdf != b"none":
        raise Exception("passphrase encrypted private keys not supported")

    if nkeys != 1:
        raise Exception(
            "multiple private keys in a key file not supported (found {} keys)".format(
                nkeys
            )
        )

    if mac:
        raise Exception("invalid OpenSSH private key (found remaining payload for mac)")

    packet = _SSHPacketReader(key_data)

    packet.get_uint32()  # check1
    packet.get_uint32()  # check2

    alg = packet.get_string()

    if alg != b"ssh-ed25519":
        raise Exception(
            'invalid key type: we only support Ed25519 (found "{}")'.format(
                alg.decode("ascii")
            )
        )

    vk = packet.get_string()
    sk = packet.get_string()

    if len(vk) != bindings.crypto_sign_PUBLICKEYBYTES:
        raise Exception("invalid public key length")

    if len(sk) != bindings.crypto_sign_SECRETKEYBYTES:
        raise Exception("invalid public key length")

    comment = packet.get_string()  # comment
    pad = packet.get_remaining_payload()

    if len(pad) and (len(pad) >= block_size or pad != _makepad(len(pad))):
        raise Exception(
            "invalid OpenSSH private key (padlen={}, actual_pad={}, expected_pad={})".format(
                len(pad), pad, _makepad(len(pad))
            )
        )

    # secret key (64 octets) = 32 octets seed || 32 octets secret key derived of seed
    seed = sk[: bindings.crypto_sign_SEEDBYTES]

    comment = comment.decode("ascii")

    return seed, comment


def _read_signify_ed25519_signature(signature_file):
    """
    Read a Ed25519 signature file created with OpenBSD signify.

    http://man.openbsd.org/OpenBSD-current/man1/signify.1
    """
    with open(signature_file) as f:
        # signature file format: 2nd line is base64 of 'Ed' || 8 random octets || 64 octets Ed25519 signature
        sig = binascii.a2b_base64(f.read().splitlines()[1])[10:]
        if len(sig) != 64:
            raise Exception(
                "bogus Ed25519 signature: raw signature length was {}, but expected 64".format(
                    len(sig)
                )
            )
        return sig


def _read_signify_ed25519_pubkey(pubkey_file):
    """
    Read a public key from a Ed25519 key pair created with OpenBSD signify.

    http://man.openbsd.org/OpenBSD-current/man1/signify.1
    """
    with open(pubkey_file) as f:
        # signature file format: 2nd line is base64 of 'Ed' || 8 random octets || 32 octets Ed25519 public key
        pubkey = binascii.a2b_base64(f.read().splitlines()[1])[10:]
        if len(pubkey) != 32:
            raise Exception(
                "bogus Ed25519 public key: raw key length was {}, but expected 32".format(
                    len(pubkey)
                )
            )
        return pubkey


def _qrcode_from_signify_ed25519_pubkey(pubkey_file, mode="text"):
    """

    Usage:

    1. Get the OpenBSD 5.7 release public key from here

        http://cvsweb.openbsd.org/cgi-bin/cvsweb/src/etc/signify/Attic/openbsd-57-base.pub?rev=1.1

    2. Generate QR Code and print to terminal

        print(cryptosign._qrcode_from_signify_ed25519_pubkey('openbsd-57-base.pub'))

    3. Compare to (scroll down) QR code here

        https://www.openbsd.org/papers/bsdcan-signify.html
    """
    assert mode in ["text", "svg"]

    import qrcode

    with open(pubkey_file) as f:
        pubkey = f.read().splitlines()[1]

        qr = qrcode.QRCode(box_size=3, error_correction=qrcode.ERROR_CORRECT_L)
        qr.add_data(pubkey)

        if mode == "text":
            import io

            with io.StringIO() as data_buffer:
                qr.print_ascii(out=data_buffer, invert=True)
                return data_buffer.getvalue()
        elif mode == "svg":
            import qrcode.image.svg

            image = qr.make_image(image_factory=qrcode.image.svg.SvgImage)
            return image.to_string()
        else:
            raise Exception("logic error")


def _verify_signify_ed25519_signature(pubkey_file, signature_file, message):
    """
    Verify a Ed25519 signature created with OpenBSD signify.

    This will raise a `nacl.exceptions.BadSignatureError` if the signature is bad
    and return silently when the signature is good.

    Usage:

    1. Create a signature:

        signify-openbsd -S -s ~/.signify/crossbario-trustroot.sec -m .profile

    2. Verify the signature

        from autobahn.wamp import cryptosign

        with open('.profile', 'rb') as f:
            message = f.read()
            cryptosign._verify_signify_ed25519_signature('.signify/crossbario-trustroot.pub', '.profile.sig', message)

    http://man.openbsd.org/OpenBSD-current/man1/signify.1
    """
    pubkey = _read_signify_ed25519_pubkey(pubkey_file)
    verify_key = signing.VerifyKey(pubkey)
    sig = _read_signify_ed25519_signature(signature_file)
    verify_key.verify(message, sig)


# CryptosignKey from
#   - raw byte string or file with raw bytes
#   - SSH private key string or key file
#   - SSH agent proxy
#
# VerifyKey from
#   - raw byte string or file with raw bytes
#   - SSH public key string or key file

if HAS_CRYPTOSIGN:

    def _format_challenge(
        challenge: Challenge,
        channel_id_raw: Optional[bytes],
        channel_id_type: Optional[str],
    ) -> bytes:
        """
        Format the challenge based on provided parameters

        :param challenge: The WAMP-cryptosign challenge object for which a signature should be computed.
        :param channel_id_raw: The channel ID when channel_id_type is 'tls-unique'.
        :param channel_id_type: The type of the channel id, currently handles 'tls-unique' and
            ignores otherwise.
        """
        if not isinstance(challenge, Challenge):
            raise Exception(
                "challenge must be instance of autobahn.wamp.types.Challenge, not {}".format(
                    type(challenge)
                )
            )

        if "challenge" not in challenge.extra:
            raise Exception("missing challenge value in challenge.extra")

        # the challenge sent by the router (a 32 bytes random value)
        challenge_hex = challenge.extra["challenge"]

        if type(challenge_hex) != str:
            raise Exception(
                "invalid type {} for challenge (expected a hex string)".format(
                    type(challenge_hex)
                )
            )

        if len(challenge_hex) != 64:
            raise Exception(
                "unexpected challenge (hex) length: was {}, but expected 64".format(
                    len(challenge_hex)
                )
            )

        # the challenge for WAMP-cryptosign is a 32 bytes random value in Hex encoding (that is, a unicode string)
        challenge_raw = binascii.a2b_hex(challenge_hex)

        if channel_id_type == "tls-unique":
            assert len(channel_id_raw) == 32, (
                "unexpected TLS transport channel ID length (was {}, but expected 32)".format(
                    len(channel_id_raw)
                )
            )

            # with TLS channel binding of type "tls-unique", the message to be signed by the client actually
            # is the XOR of the challenge and the TLS channel ID
            data = util.xor(challenge_raw, channel_id_raw)
        elif channel_id_type is None:
            # when no channel binding was requested, the message to be signed by the client is the challenge only
            data = challenge_raw
        else:
            assert False, 'invalid channel_id_type "{}"'.format(channel_id_type)

        return data

    def _sign_challenge(data: bytes, signer_func: Callable) -> bytes:
        """
        Sign the provided data using the provided signer.

        :param data: challenge to sign
        :param signer_func: The callable function to use for signing
        :returns: A Deferred/Future that resolves to the computed signature.
        :rtype: str
        """
        # a raw byte string is signed, and the signature is also a raw byte string
        d1 = signer_func(data)

        # asyncio lacks callback chaining (and we cannot use co-routines, since we want
        # to support older Pythons), hence we need d2
        d2 = txaio.create_future()

        def process(signature_raw):
            # convert the raw signature into a hex encode value (unicode string)
            signature_hex = binascii.b2a_hex(signature_raw).decode("ascii")

            # we return the concatenation of the signature and the message signed (96 bytes)
            data_hex = binascii.b2a_hex(data).decode("ascii")

            sig = signature_hex + data_hex
            txaio.resolve(d2, sig)

        txaio.add_callbacks(d1, process, None)

        return d2

    class CryptosignKey(object):
        """
        A cryptosign private key for signing, and hence usable for authentication or a
        public key usable for verification (but can't be used for signing).
        """

        def __init__(
            self,
            key,
            can_sign: bool,
            security_module: Optional[ISecurityModule] = None,
            key_no: Optional[int] = None,
            comment: Optional[str] = None,
        ) -> None:
            if not (
                isinstance(key, signing.VerifyKey)
                or isinstance(key, signing.SigningKey)
            ):
                raise Exception("invalid type {} for key".format(type(key)))

            assert (can_sign and isinstance(key, signing.SigningKey)) or (
                not can_sign and isinstance(key, signing.VerifyKey)
            )
            self._key = key
            self._can_sign = can_sign
            self._security_module = security_module
            self._key_no = key_no
            self._comment = comment

        @property
        def security_module(self) -> Optional["ISecurityModule"]:
            """
            Implements :meth:`autobahn.wamp.interfaces.IKey.security_module`.
            """
            return self._security_module

        @property
        def key_no(self) -> Optional[int]:
            """
            Implements :meth:`autobahn.wamp.interfaces.IKey.key_no`.
            """
            return self._key_no

        @property
        def comment(self) -> Optional[str]:
            """
            Implements :meth:`autobahn.wamp.interfaces.IKey.comment`.
            """
            return self._comment

        @property
        def key_type(self) -> str:
            """
            Implements :meth:`autobahn.wamp.interfaces.IKey.key_type`.
            """
            return "cryptosign"

        @property
        def can_sign(self) -> bool:
            """
            Implements :meth:`autobahn.wamp.interfaces.IKey.can_sign`.
            """
            return self._can_sign

        def sign(self, data: bytes) -> bytes:
            """
            Implements :meth:`autobahn.wamp.interfaces.IKey.sign`.
            """
            if not self._can_sign:
                raise Exception("a signing key required to sign")

            if type(data) != bytes:
                raise Exception("data to be signed must be binary")

            sig: SignedMessage = self._key.sign(data)

            # we only return the actual signature! if we return "sig",
            # it gets coerced into the concatenation of message + signature
            # not sure which order, but we don't want that. we only want
            # the signature
            return txaio.create_future_success(sig.signature)

        def sign_challenge(
            self,
            challenge: Challenge,
            channel_id: Optional[bytes] = None,
            channel_id_type: Optional[str] = None,
        ) -> bytes:
            """
            Implements :meth:`autobahn.wamp.interfaces.ICryptosignKey.sign_challenge`.
            """
            assert challenge.method in [
                "cryptosign",
                "cryptosign-proxy",
            ], 'unexpected cryptosign challenge with method "{}"'.format(
                challenge.method
            )

            data = _format_challenge(challenge, channel_id, channel_id_type)

            return _sign_challenge(data, self.sign)

        def public_key(self, binary: bool = False) -> Union[str, bytes]:
            """
            Returns the public key part of a signing key or the (public) verification key.

            :returns: The public key in Hex encoding.
            :rtype: str or None
            """
            if isinstance(self._key, signing.SigningKey):
                key = self._key.verify_key
            else:
                key = self._key

            if binary:
                return key.encode()
            else:
                return key.encode(encoder=encoding.HexEncoder).decode("ascii")

        @classmethod
        def from_pubkey(
            cls, pubkey: bytes, comment: Optional[str] = None
        ) -> "CryptosignKey":
            if not (comment is None or type(comment) == str):
                raise ValueError("invalid type {} for comment".format(type(comment)))

            if type(pubkey) != bytes:
                raise ValueError(
                    "invalid key type {} (expected binary)".format(type(pubkey))
                )

            if len(pubkey) != 32:
                raise ValueError(
                    "invalid key length {} (expected 32)".format(len(pubkey))
                )

            return cls(key=signing.VerifyKey(pubkey), can_sign=False, comment=comment)

        @classmethod
        def from_bytes(
            cls, key: bytes, comment: Optional[str] = None
        ) -> "CryptosignKey":
            if not (comment is None or type(comment) == str):
                raise ValueError("invalid type {} for comment".format(type(comment)))

            if type(key) != bytes:
                raise ValueError(
                    "invalid key type {} (expected binary)".format(type(key))
                )

            if len(key) != 32:
                raise ValueError("invalid key length {} (expected 32)".format(len(key)))

            return cls(key=signing.SigningKey(key), can_sign=True, comment=comment)

        @classmethod
        def from_file(
            cls, filename: str, comment: Optional[str] = None
        ) -> "CryptosignKey":
            """
            Load an Ed25519 (private) signing key (actually, the seed for the key) from a raw file of 32 bytes length.
            This can be any random byte sequence, such as generated from Python code like

                os.urandom(32)

            or from the shell

                dd if=/dev/urandom of=client02.key bs=1 count=32

            :param filename: Filename of the key.
            :param comment: Comment for key (optional).
            """
            if not (comment is None or type(comment) == str):
                raise Exception("invalid type {} for comment".format(type(comment)))

            if type(filename) != str:
                raise Exception("invalid type {} for filename".format(filename))

            with open(filename, "rb") as f:
                key_data = f.read()

            return cls.from_bytes(key_data, comment=comment)

        @classmethod
        def from_ssh_file(cls, filename: str) -> "CryptosignKey":
            """
            Load an Ed25519 key from a SSH key file. The key file can be a (private) signing
            key (from a SSH private key file) or a (public) verification key (from a SSH
            public key file). A private key file must be passphrase-less.
            """

            with open(filename, "rb") as f:
                key_data = f.read().decode("utf-8").strip()
            return cls.from_ssh_bytes(key_data)

        @classmethod
        def from_ssh_bytes(cls, key_data: str) -> "CryptosignKey":
            """
            Load an Ed25519 key from SSH key file. The key file can be a (private) signing
            key (from a SSH private key file) or a (public) verification key (from a SSH
            public key file). A private key file must be passphrase-less.
            """
            SSH_BEGIN = "-----BEGIN OPENSSH PRIVATE KEY-----"
            if key_data.startswith(SSH_BEGIN):
                # OpenSSH private key
                key_data, comment = _read_ssh_ed25519_privkey(key_data)
                key = signing.SigningKey(key_data, encoder=encoding.RawEncoder)
                can_sign = True
            else:
                # OpenSSH public key
                key_data, comment = _read_ssh_ed25519_pubkey(key_data)
                key = signing.VerifyKey(key_data)
                can_sign = False

            return cls(key=key, can_sign=can_sign, comment=comment)

        @classmethod
        def from_seedphrase(cls, seedphrase: str, index: int = 0) -> "CryptosignKey":
            """
            Create a private key from the given BIP-39 mnemonic seed phrase and index,
            which can be used to sign and create signatures.

            :param seedphrase: The BIP-39 seedphrase ("Mnemonic") from which to derive the account.
            :param index: The account index in account hierarchy defined by the seedphrase.
            :return: New instance of :class:`EthereumKey`
            """
            # BIP44 path for WAMP
            # https://github.com/wamp-proto/wamp-proto/issues/401
            # https://github.com/satoshilabs/slips/pull/1322
            derivation_path = "m/44'/655'/0'/0/{}".format(index)

            key_raw = mnemonic_to_private_key(seedphrase, derivation_path)
            assert type(key_raw) == bytes
            assert len(key_raw) == 32

            # create WAMP-Cryptosign key object from raw bytes
            key = cls.from_bytes(key_raw)

            return key

        @classmethod
        def from_keyfile(cls, keyfile: str) -> "CryptosignKey":
            """
            Create a public or private key from reading the given public or private key file.

            Here is an example key file that includes an CryptosignKey private key ``private-key-ed25519``, which
            is loaded in this function, and other fields, which are ignored by this function:

            .. code-block::

                This is a comment (all lines until the first empty line are comments indeed).

                creator: oberstet@intel-nuci7
                created-at: 2022-07-05T12:29:48.832Z
                user-id: oberstet@intel-nuci7
                public-key-ed25519: 7326d9dc0307681cc6940fde0e60eb31a6e4d642a81e55c434462ce31f95deed
                public-adr-eth: 0x10848feBdf7f200Ba989CDf7E3eEB3EC03ae7768
                private-key-ed25519: f750f42b0430e28a2e272c3cedcae4dcc4a1cf33bc345c35099d3322626ab666
                private-key-eth: 4d787714dcb0ae52e1c5d2144648c255d660b9a55eac9deeb80d9f506f501025

            :param keyfile: Path (relative or absolute) to a public or private keys file.
            :return: New instance of :class:`CryptosignKey`
            """
            if not os.path.exists(keyfile) or not os.path.isfile(keyfile):
                raise RuntimeError('keyfile "{}" is not a file'.format(keyfile))

            # now load the private or public key file - this returns a dict which should
            # include (for a private key):
            #
            #   private-key-ed25519: 20e8c05d0ede9506462bb049c4843032b18e8e75b314583d0c8d8a4942f9be40
            #
            # or (for a public key only):
            #
            #   public-key-ed25519: 7326d9dc0307681cc6940fde0e60eb31a6e4d642a81e55c434462ce31f95deed
            #
            data = parse_keyfile(keyfile)

            privkey_ed25519_hex = data.get("private-key-ed25519", None)
            if privkey_ed25519_hex is None:
                pubkey_ed25519_hex = data.get("public-key-ed25519", None)
                if pubkey_ed25519_hex is None:
                    raise RuntimeError(
                        'neither "private-key-ed25519" nor "public-key-ed25519" found in keyfile {}'.format(
                            keyfile
                        )
                    )
                else:
                    return CryptosignKey.from_pubkey(
                        binascii.a2b_hex(pubkey_ed25519_hex)
                    )
            else:
                return CryptosignKey.from_bytes(binascii.a2b_hex(privkey_ed25519_hex))

    ICryptosignKey.register(CryptosignKey)

    class CryptosignAuthextra(object):
        """
        WAMP-Cryptosign authextra object.
        """

        __slots__ = [
            "_pubkey",
            "_trustroot",
            "_challenge",
            "_channel_binding",
            "_channel_id",
            "_realm",
            "_chain_id",
            "_block_no",
            "_delegate",
            "_seeder",
            "_bandwidth",
            "_signature",
        ]

        def __init__(
            self,
            pubkey: Optional[bytes] = None,
            challenge: Optional[bytes] = None,
            channel_binding: Optional[str] = None,
            channel_id: Optional[bytes] = None,
            # domain address, certificates are verified against owner of the domain
            trustroot: Optional[bytes] = None,
            # FIXME: add delegate address
            # FIXME: add certificates
            # FIXME: remove reservation
            realm: Optional[bytes] = None,
            chain_id: Optional[int] = None,
            block_no: Optional[int] = None,
            delegate: Optional[bytes] = None,
            seeder: Optional[bytes] = None,
            bandwidth: Optional[int] = None,
            signature: Optional[bytes] = None,
        ):
            if pubkey:
                assert len(pubkey) == 32
            if trustroot:
                assert len(trustroot) == 20
            if challenge:
                assert len(challenge) == 32
            if channel_binding:
                assert channel_binding in ["tls-unique"]
            if channel_id:
                assert len(channel_id) == 32
            if realm:
                assert len(realm) == 20
            if delegate:
                assert len(delegate) == 20
            if seeder:
                assert len(seeder) == 20
            if signature:
                assert len(signature) == 65
            self._pubkey = pubkey
            self._trustroot = trustroot
            self._challenge = challenge
            self._channel_binding = channel_binding
            self._channel_id = channel_id
            self._realm = realm
            self._chain_id = chain_id
            self._block_no = block_no
            self._delegate = delegate
            self._seeder = seeder
            self._bandwidth = bandwidth
            self._signature = signature

        @property
        def pubkey(self) -> Optional[bytes]:
            return self._pubkey

        @pubkey.setter
        def pubkey(self, value: Optional[bytes]):
            assert value is None or len(value) == 20
            self._pubkey = value

        @property
        def trustroot(self) -> Optional[bytes]:
            return self._trustroot

        @trustroot.setter
        def trustroot(self, value: Optional[bytes]):
            assert value is None or len(value) == 20
            self._trustroot = value

        @property
        def challenge(self) -> Optional[bytes]:
            return self._challenge

        @challenge.setter
        def challenge(self, value: Optional[bytes]):
            assert value is None or len(value) == 32
            self._challenge = value

        @property
        def channel_binding(self) -> Optional[str]:
            return self._channel_binding

        @channel_binding.setter
        def channel_binding(self, value: Optional[str]):
            assert value is None or value in ["tls-unique"]
            self._channel_binding = value

        @property
        def channel_id(self) -> Optional[bytes]:
            return self._channel_id

        @channel_id.setter
        def channel_id(self, value: Optional[bytes]):
            assert value is None or len(value) == 32
            self._channel_id = value

        @property
        def realm(self) -> Optional[bytes]:
            return self._realm

        @realm.setter
        def realm(self, value: Optional[bytes]):
            assert value is None or len(value) == 20
            self._realm = value

        @property
        def chain_id(self) -> Optional[int]:
            return self._chain_id

        @chain_id.setter
        def chain_id(self, value: Optional[int]):
            assert value is None or value > 0
            self._chain_id = value

        @property
        def block_no(self) -> Optional[int]:
            return self._block_no

        @block_no.setter
        def block_no(self, value: Optional[int]):
            assert value is None or value > 0
            self._block_no = value

        @property
        def delegate(self) -> Optional[bytes]:
            return self._delegate

        @delegate.setter
        def delegate(self, value: Optional[bytes]):
            assert value is None or len(value) == 20
            self._delegate = value

        @property
        def seeder(self) -> Optional[bytes]:
            return self._seeder

        @seeder.setter
        def seeder(self, value: Optional[bytes]):
            assert value is None or len(value) == 20
            self._seeder = value

        @property
        def bandwidth(self) -> Optional[int]:
            return self._bandwidth

        @bandwidth.setter
        def bandwidth(self, value: Optional[int]):
            assert value is None or value > 0
            self._bandwidth = value

        @property
        def signature(self) -> Optional[bytes]:
            return self._signature

        @signature.setter
        def signature(self, value: Optional[bytes]):
            assert value is None or len(value) == 65
            self._signature = value

        @staticmethod
        def parse(data: Dict[str, Any]) -> "CryptosignAuthextra":
            obj = CryptosignAuthextra()

            pubkey = data.get("pubkey", None)
            if pubkey is not None:
                if type(pubkey) != str:
                    raise ValueError("invalid type {} for pubkey".format(type(pubkey)))
                if len(pubkey) != 32 * 2:
                    raise ValueError("invalid length {} of pubkey".format(len(pubkey)))
                obj._pubkey = a2b_hex(pubkey)

            challenge = data.get("challenge", None)
            if challenge is not None:
                if type(challenge) != str:
                    raise ValueError(
                        "invalid type {} for challenge".format(type(challenge))
                    )
                if len(challenge) != 32 * 2:
                    raise ValueError(
                        "invalid length {} of challenge".format(len(challenge))
                    )
                obj._challenge = a2b_hex(challenge)

            channel_binding = data.get("channel_binding", None)
            if channel_binding is not None:
                if type(channel_binding) != str:
                    raise ValueError(
                        "invalid type {} for channel_binding".format(
                            type(channel_binding)
                        )
                    )
                if channel_binding not in ["tls-unique"]:
                    raise ValueError(
                        'invalid value "{}" for channel_binding'.format(channel_binding)
                    )
                obj._channel_binding = channel_binding

            channel_id = data.get("channel_id", None)
            if channel_id is not None:
                if type(channel_id) != str:
                    raise ValueError(
                        "invalid type {} for channel_id".format(type(channel_id))
                    )
                if len(channel_id) != 32 * 2:
                    raise ValueError(
                        "invalid length {} of channel_id".format(len(channel_id))
                    )
                obj._channel_id = a2b_hex(channel_id)

            trustroot = data.get("trustroot", None)
            if trustroot is not None:
                if type(trustroot) != str:
                    raise ValueError(
                        "invalid type {} for trustroot - expected a string".format(
                            type(trustroot)
                        )
                    )
                if not _URI_PAT_REALM_NAME_ETH.match(trustroot):
                    raise ValueError(
                        'invalid value "{}" for trustroot - expected an Ethereum address'.format(
                            type(trustroot)
                        )
                    )
                obj._trustroot = a2b_hex(trustroot[2:])

            reservation = data.get("reservation", None)
            if reservation is not None:
                if type(reservation) != dict:
                    raise ValueError(
                        "invalid type {} for reservation".format(type(reservation))
                    )

                chain_id = reservation.get("chain_id", None)
                if chain_id is not None:
                    if type(chain_id) != int:
                        raise ValueError(
                            "invalid type {} for reservation.chain_id - expected an integer".format(
                                type(chain_id)
                            )
                        )
                    obj._chain_id = chain_id

                block_no = reservation.get("block_no", None)
                if block_no is not None:
                    if type(block_no) != int:
                        raise ValueError(
                            "invalid type {} for reservation.block_no - expected an integer".format(
                                type(block_no)
                            )
                        )
                    obj._block_no = block_no

                realm = reservation.get("realm", None)
                if realm is not None:
                    if type(realm) != str:
                        raise ValueError(
                            "invalid type {} for reservation.realm - expected a string".format(
                                type(realm)
                            )
                        )
                    if not _URI_PAT_REALM_NAME_ETH.match(realm):
                        raise ValueError(
                            'invalid value "{}" for reservation.realm - expected an Ethereum address'.format(
                                type(realm)
                            )
                        )
                    obj._realm = a2b_hex(realm[2:])

                delegate = reservation.get("delegate", None)
                if delegate is not None:
                    if type(delegate) != str:
                        raise ValueError(
                            "invalid type {} for reservation.delegate - expected a string".format(
                                type(delegate)
                            )
                        )
                    if not _URI_PAT_REALM_NAME_ETH.match(delegate):
                        raise ValueError(
                            'invalid value "{}" for reservation.delegate - expected an Ethereum address'.format(
                                type(delegate)
                            )
                        )
                    obj._delegate = a2b_hex(delegate[2:])

                seeder = reservation.get("seeder", None)
                if seeder is not None:
                    if type(seeder) != str:
                        raise ValueError(
                            "invalid type {} for reservation.seeder - expected a string".format(
                                type(seeder)
                            )
                        )
                    if not _URI_PAT_REALM_NAME_ETH.match(seeder):
                        raise ValueError(
                            'invalid value "{}" for reservation.seeder - expected an Ethereum address'.format(
                                type(seeder)
                            )
                        )
                    obj._seeder = a2b_hex(seeder[2:])

                bandwidth = reservation.get("bandwidth", None)
                if bandwidth is not None:
                    if type(bandwidth) != int:
                        raise ValueError(
                            "invalid type {} for reservation.bandwidth - expected an integer".format(
                                type(bandwidth)
                            )
                        )
                    obj._bandwidth = bandwidth

            signature = data.get("signature", None)
            if signature is not None:
                if type(signature) != str:
                    raise ValueError(
                        "invalid type {} for signature".format(type(signature))
                    )
                if len(signature) != 65 * 2:
                    raise ValueError(
                        "invalid length {} of signature".format(len(signature))
                    )
                obj._signature = a2b_hex(signature)

            return obj

        def marshal(self) -> Dict[str, Any]:
            res = {}

            # FIXME: marshal check-summed eth addresses

            if self._pubkey is not None:
                res["pubkey"] = b2a_hex(self._pubkey).decode()

            if self._challenge is not None:
                res["challenge"] = b2a_hex(self._challenge).decode()
            if self._channel_binding is not None:
                res["channel_binding"] = self._channel_binding
            if self._channel_id is not None:
                res["channel_id"] = b2a_hex(self._channel_id).decode()

            if self._trustroot is not None:
                res["trustroot"] = "0x" + b2a_hex(self._trustroot).decode()

            reservation = {}
            if self._chain_id is not None:
                reservation["chain_id"] = self._chain_id
            if self._block_no is not None:
                reservation["block_no"] = self._block_no
            if self._realm is not None:
                reservation["realm"] = "0x" + b2a_hex(self._realm).decode()
            if self._delegate is not None:
                reservation["delegate"] = "0x" + b2a_hex(self._delegate).decode()
            if self._seeder is not None:
                reservation["seeder"] = "0x" + b2a_hex(self._seeder).decode()
            if self._bandwidth is not None:
                reservation["bandwidth"] = self._bandwidth
            if reservation:
                res["reservation"] = reservation

            if self._signature is not None:
                res["signature"] = b2a_hex(self._signature).decode()

            return res

    __all__.extend(
        ["CryptosignAuthextra", "CryptosignKey", "format_challenge", "sign_challenge"]
    )



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import base64
import binascii
import hashlib
import hmac
import os
import random
import struct
import time
from typing import Dict, Optional

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from autobahn.util import public
from autobahn.util import xor as xor_array
from autobahn.wamp.interfaces import IAuthenticator

# if we don't have argon2/passlib (see "authentication" extra) then
# you don't get AuthScram and variants
try:
    from argon2 import Type
    from argon2.low_level import hash_secret
    from passlib.utils import saslprep

    HAS_ARGON = True
except ImportError:
    HAS_ARGON = False


__all__ = (
    "AuthAnonymous",
    "AuthCryptoSign",
    "AuthScram",
    "AuthTicket",
    "AuthWampCra",
    "check_totp",
    "compute_totp",
    "compute_wcs",
    "create_authenticator",
    "derive_key",
    "derive_scram_credential",
    "generate_totp_secret",
    "generate_wcs",
    "pbkdf2",
    "qrcode_from_totp",
)


def create_authenticator(name, **kwargs):
    """
    Accepts various keys and values to configure an authenticator. The
    valid keys depend on the kind of authenticator but all can
    understand: `authextra`, `authid` and `authrole`

    :return: an instance implementing IAuthenticator with the given
        configuration.
    """
    try:
        klass = {
            AuthScram.name: AuthScram,
            AuthCryptoSign.name: AuthCryptoSign,
            AuthCryptoSignProxy.name: AuthCryptoSignProxy,
            AuthWampCra.name: AuthWampCra,
            AuthAnonymous.name: AuthAnonymous,
            AuthAnonymousProxy.name: AuthAnonymousProxy,
            AuthTicket.name: AuthTicket,
        }[name]
    except KeyError:
        raise ValueError("Unknown authenticator '{}'".format(name))
    # this may raise further ValueErrors if the kwargs are wrong
    authenticator = klass(**kwargs)
    return authenticator


# experimental authentication API
class AuthAnonymous(object):
    name = "anonymous"

    def __init__(self, **kw):
        self._args = kw

    @property
    def authextra(self):
        return self._args.get("authextra", dict())

    def on_challenge(self, session, challenge):
        raise RuntimeError("on_challenge called on anonymous authentication")

    def on_welcome(self, msg, authextra):
        return None


IAuthenticator.register(AuthAnonymous)


class AuthAnonymousProxy(AuthAnonymous):
    name = "anonymous-proxy"


IAuthenticator.register(AuthAnonymousProxy)


class AuthTicket(object):
    name = "ticket"

    def __init__(self, **kw):
        self._args = kw
        try:
            self._ticket = self._args.pop("ticket")
        except KeyError:
            raise ValueError("ticket authentication requires 'ticket=' kwarg")

    @property
    def authextra(self):
        return self._args.get("authextra", dict())

    def on_challenge(self, session, challenge):
        assert challenge.method == "ticket"
        return self._ticket

    def on_welcome(self, msg, authextra):
        return None


IAuthenticator.register(AuthTicket)


class AuthCryptoSign(object):
    name = "cryptosign"

    def __init__(self, **kw):
        # should put in checkconfig or similar
        for key in kw.keys():
            if key not in ["authextra", "authid", "authrole", "privkey"]:
                raise ValueError(
                    "Unexpected key '{}' for {}".format(key, self.__class__.__name__)
                )
        for key in ["privkey"]:
            if key not in kw:
                raise ValueError("Must provide '{}' for cryptosign".format(key))

        from autobahn.wamp.cryptosign import CryptosignKey

        self._privkey = CryptosignKey.from_bytes(binascii.a2b_hex(kw["privkey"]))

        if "pubkey" in kw.get("authextra", dict()):
            pubkey = kw["authextra"]["pubkey"]
            if pubkey != self._privkey.public_key():
                raise ValueError("Public key doesn't correspond to private key")
        else:
            kw["authextra"] = kw.get("authextra", dict())
            kw["authextra"]["pubkey"] = self._privkey.public_key()

        self._channel_binding = kw.get("authextra", dict()).get("channel_binding", None)
        self._args = kw

    @property
    def authextra(self):
        return self._args.get("authextra", dict())

    def on_challenge(self, session, challenge):
        channel_id = session._transport.transport_details.channel_id.get(
            self._channel_binding, None
        )
        return self._privkey.sign_challenge(
            challenge, channel_id=channel_id, channel_id_type=self._channel_binding
        )

    def on_welcome(self, msg, authextra):
        return None


IAuthenticator.register(AuthCryptoSign)


class AuthCryptoSignProxy(AuthCryptoSign):
    name = "cryptosign-proxy"


IAuthenticator.register(AuthCryptoSignProxy)


def _hash_argon2id13_secret(password, salt, iterations, memory):
    """
    Internal helper. Returns the salted/hashed password using the
    argon2id-13 algorithm. The return value is base64-encoded.
    """
    rawhash = hash_secret(
        secret=password,
        salt=base64.b64decode(salt),
        time_cost=iterations,
        memory_cost=memory,
        parallelism=1,  # hard-coded by WAMP-SCRAM spec
        hash_len=32,
        type=Type.ID,
        version=0x13,  # note this is decimal "19" which appears in places
    )
    # spits out stuff like:
    # '$argon2i$v=19$m=512,t=2,p=2$5VtWOO3cGWYQHEMaYGbsfQ$AcmqasQgW/wI6wAHAMk4aQ'

    _, tag, ver, options, salt_data, hash_data = rawhash.split(b"$")
    return hash_data


def _hash_pbkdf2_secret(password, salt, iterations):
    """
    Internal helper for SCRAM authentication
    """
    return pbkdf2(password, salt, iterations, keylen=32)


class AuthScram(object):
    """
    Implements "wamp-scram" authentication for components.

    NOTE: This is a prototype of a draft spec; see
    https://github.com/wamp-proto/wamp-proto/issues/135
    """

    name = "scram"

    def __init__(self, **kw):
        if not HAS_ARGON:
            raise RuntimeError(
                "Cannot support WAMP-SCRAM without argon2_cffi and "
                "passlib libraries; install autobahn['scram']"
            )
        self._args = kw
        self._client_nonce = None

    @property
    def authextra(self):
        # is authextra() called exactly once per authentication?
        if self._client_nonce is None:
            self._client_nonce = base64.b64encode(os.urandom(16)).decode("ascii")
        return {
            "nonce": self._client_nonce,
        }

    def on_challenge(self, session, challenge):
        assert challenge.method == "scram"
        assert self._client_nonce is not None
        required_args = ["nonce", "kdf", "salt", "iterations"]
        optional_args = ["memory", "channel_binding"]
        for k in required_args:
            if k not in challenge.extra:
                raise RuntimeError(
                    "WAMP-SCRAM challenge option '{}' is "
                    " required but not specified".format(k)
                )
        for k in challenge.extra:
            if k not in optional_args + required_args:
                raise RuntimeError(
                    "WAMP-SCRAM challenge has unknown attribute '{}'".format(k)
                )

        channel_binding = challenge.extra.get("channel_binding", "")
        server_nonce = challenge.extra["nonce"]  # base64
        salt = challenge.extra["salt"]  # base64
        iterations = int(challenge.extra["iterations"])
        memory = int(challenge.extra.get("memory", -1))
        password = self._args["password"].encode("utf8")  # supplied by user
        authid = saslprep(self._args["authid"])
        algorithm = challenge.extra["kdf"]
        client_nonce = self._client_nonce

        self._auth_message = (
            "{client_first_bare},{server_first},{client_final_no_proof}".format(
                client_first_bare="n={},r={}".format(authid, client_nonce),
                server_first="r={},s={},i={}".format(server_nonce, salt, iterations),
                client_final_no_proof="c={},r={}".format(channel_binding, server_nonce),
            )
        ).encode("ascii")

        if algorithm == "argon2id-13":
            if memory == -1:
                raise ValueError(
                    "WAMP-SCRAM 'argon2id-13' challenge requires 'memory' parameter"
                )
            self._salted_password = _hash_argon2id13_secret(
                password, salt, iterations, memory
            )
        elif algorithm == "pbkdf2":
            self._salted_password = _hash_pbkdf2_secret(password, salt, iterations)
        else:
            raise RuntimeError(
                "WAMP-SCRAM specified unknown KDF '{}'".format(algorithm)
            )

        client_key = hmac.new(
            self._salted_password, b"Client Key", hashlib.sha256
        ).digest()
        stored_key = hashlib.new("sha256", client_key).digest()

        client_signature = hmac.new(
            stored_key, self._auth_message, hashlib.sha256
        ).digest()
        client_proof = xor_array(client_key, client_signature)

        return base64.b64encode(client_proof)

    def on_welcome(self, session, authextra):
        """
        When the server is satisfied, it sends a 'WELCOME' message.

        This hook allows us an opportunity to deny the session right
        before it gets set up -- we check the server-signature thus
        authorizing the server and if it fails we drop the connection.
        """
        alleged_server_sig = base64.b64decode(authextra["scram_server_signature"])
        server_key = hmac.new(
            self._salted_password, b"Server Key", hashlib.sha256
        ).digest()
        server_signature = hmac.new(
            server_key, self._auth_message, hashlib.sha256
        ).digest()
        if not hmac.compare_digest(server_signature, alleged_server_sig):
            session.log.error("Verification of server SCRAM signature failed")
            return "Verification of server SCRAM signature failed"
        session.log.info("Verification of server SCRAM signature successful")
        return None


IAuthenticator.register(AuthScram)


class AuthWampCra(object):
    name = "wampcra"

    def __init__(self, **kw):
        # should put in checkconfig or similar
        for key in kw.keys():
            if key not in ["authextra", "authid", "authrole", "secret"]:
                raise ValueError(
                    "Unexpected key '{}' for {}".format(key, self.__class__.__name__)
                )
        for key in ["secret", "authid"]:
            if key not in kw:
                raise ValueError("Must provide '{}' for wampcra".format(key))

        self._args = kw
        self._secret = kw.pop("secret")
        if not isinstance(self._secret, str):
            self._secret = self._secret.decode("utf8")

    @property
    def authextra(self):
        return self._args.get("authextra", dict())

    def on_challenge(self, session, challenge):
        key = self._secret.encode("utf8")
        if "salt" in challenge.extra:
            key = derive_key(
                key,
                challenge.extra["salt"],
                challenge.extra["iterations"],
                challenge.extra["keylen"],
            )

        signature = compute_wcs(key, challenge.extra["challenge"].encode("utf8"))
        return signature.decode("ascii")

    def on_welcome(self, msg, authextra):
        return None


IAuthenticator.register(AuthWampCra)


@public
def generate_totp_secret(length=10):
    """
    Generates a new Base32 encoded, random secret.

    .. seealso:: http://en.wikipedia.org/wiki/Base32

    :param length: The length of the entropy used to generate the secret.
    :type length: int

    :returns: The generated secret in Base32 (letters ``A-Z`` and digits ``2-7``).
       The length of the generated secret is ``length * 8 / 5`` octets.
    :rtype: unicode
    """
    assert type(length) == int
    return base64.b32encode(os.urandom(length)).decode("ascii")


@public
def compute_totp(secret, offset=0):
    """
    Computes the current TOTP code.

    :param secret: Base32 encoded secret.
    :type secret: unicode
    :param offset: Time offset (in steps, use eg -1, 0, +1 for compliance with RFC6238)
        for which to compute TOTP.
    :type offset: int

    :returns: TOTP for current time (+/- offset).
    :rtype: unicode
    """
    assert type(secret) == str
    assert type(offset) == int
    try:
        key = base64.b32decode(secret)
    except TypeError:
        raise Exception("invalid secret")
    interval = offset + int(time.time()) // 30
    msg = struct.pack(">Q", interval)
    digest = hmac.new(key, msg, hashlib.sha1).digest()
    o = 15 & (digest[19])
    token = (struct.unpack(">I", digest[o : o + 4])[0] & 0x7FFFFFFF) % 1000000
    return "{0:06d}".format(token)


@public
def check_totp(secret, ticket):
    """
    Check a TOTP value received from a principal trying to authenticate against
    the expected value computed from the secret shared between the principal and
    the authenticating entity.

    The Internet can be slow, and clocks might not match exactly, so some
    leniency is allowed. RFC6238 recommends looking an extra time step in either
    direction, which essentially opens the window from 30 seconds to 90 seconds.

    :param secret: The secret shared between the principal (eg a client) that
        is authenticating, and the authenticating entity (eg a server).
    :type secret: unicode
    :param ticket: The TOTP value to be checked.
    :type ticket: unicode

    :returns: ``True`` if the TOTP value is correct, else ``False``.
    :rtype: bool
    """
    for offset in [0, 1, -1]:
        if ticket == compute_totp(secret, offset):
            return True
    return False


@public
def qrcode_from_totp(secret, label, issuer):
    if type(secret) != str:
        raise Exception("secret must be of type unicode, not {}".format(type(secret)))

    if type(label) != str:
        raise Exception("label must be of type unicode, not {}".format(type(label)))

    try:
        import qrcode
        import qrcode.image.svg
    except ImportError:
        raise Exception("qrcode not installed")

    return qrcode.make(
        "otpauth://totp/{}?secret={}&issuer={}".format(label, secret, issuer),
        box_size=3,
        image_factory=qrcode.image.svg.SvgImage,
    ).to_string()


@public
def pbkdf2(data, salt, iterations=1000, keylen=32, hashfunc=None):
    """
    Returns a binary digest for the PBKDF2 hash algorithm of ``data``
    with the given ``salt``. It iterates ``iterations`` time and produces a
    key of ``keylen`` bytes. By default SHA-256 is used as hash function,
    a different hashlib ``hashfunc`` can be provided.

    :param data: The data for which to compute the PBKDF2 derived key.
    :type data: bytes
    :param salt: The salt to use for deriving the key.
    :type salt: bytes
    :param iterations: The number of iterations to perform in PBKDF2.
    :type iterations: int
    :param keylen: The length of the cryptographic key to derive.
    :type keylen: int
    :param hashfunc: Name of the hash algorithm to use
    :type hashfunc: str

    :returns: The derived cryptographic key.
    :rtype: bytes
    """
    if (
        not (type(data) == bytes)
        or not (type(salt) == bytes)
        or not (type(iterations) == int)
        or not (type(keylen) == int)
    ):
        raise ValueError("Invalid argument types")

    # justification: WAMP-CRA uses SHA256 and users shouldn't have any
    # other reason to call this particular pbkdf2 function (arguably,
    # it should be private maybe?)
    if hashfunc is None:
        hashfunc = "sha256"
    if hashfunc is callable:
        # used to take stuff from hashlib; translate?
        raise ValueError(
            "pbkdf2 now takes the name of a hash algorithm for 'hashfunc='"
        )

    backend = default_backend()

    # https://cryptography.io/en/latest/hazmat/primitives/key-derivation-functions/#pbkdf2
    kdf = PBKDF2HMAC(
        algorithm=getattr(hashes, hashfunc.upper())(),
        length=keylen,
        salt=salt,
        iterations=iterations,
        backend=backend,
    )
    return kdf.derive(data)


@public
def derive_key(secret, salt, iterations=1000, keylen=32):
    """
    Computes a derived cryptographic key from a password according to PBKDF2.

    .. seealso:: http://en.wikipedia.org/wiki/PBKDF2

    :param secret: The secret.
    :type secret: bytes or unicode
    :param salt: The salt to be used.
    :type salt: bytes or unicode
    :param iterations: Number of iterations of derivation algorithm to run.
    :type iterations: int
    :param keylen: Length of the key to derive in bytes.
    :type keylen: int

    :return: The derived key in Base64 encoding.
    :rtype: bytes
    """
    if not (type(secret) in [str, bytes]):
        raise ValueError("'secret' must be bytes")
    if not (type(salt) in [str, bytes]):
        raise ValueError("'salt' must be bytes")
    if not (type(iterations) == int):
        raise ValueError("'iterations' must be an integer")
    if not (type(keylen) == int):
        raise ValueError("'keylen' must be an integer")
    if type(secret) == str:
        secret = secret.encode("utf8")
    if type(salt) == str:
        salt = salt.encode("utf8")
    key = pbkdf2(secret, salt, iterations, keylen)
    return binascii.b2a_base64(key).strip()


WCS_SECRET_CHARSET = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
"""
The characters from which :func:`autobahn.wamp.auth.generate_wcs` generates secrets.
"""


@public
def generate_wcs(length=14):
    """
    Generates a new random secret for use with WAMP-CRA.

    The secret generated is a random character sequence drawn from

    - upper and lower case latin letters
    - digits
    -

    :param length: The length of the secret to generate.
    :type length: int

    :return: The generated secret. The length of the generated is ``length`` octets.
    :rtype: bytes
    """
    assert type(length) == int
    return "".join(random.choice(WCS_SECRET_CHARSET) for _ in range(length)).encode(
        "ascii"
    )


@public
def compute_wcs(key, challenge):
    """
    Compute an WAMP-CRA authentication signature from an authentication
    challenge and a (derived) key.

    :param key: The key derived (via PBKDF2) from the secret.
    :type key: bytes
    :param challenge: The authentication challenge to sign.
    :type challenge: bytes

    :return: The authentication signature.
    :rtype: bytes
    """
    assert type(key) in [str, bytes]
    assert type(challenge) in [str, bytes]
    if type(key) == str:
        key = key.encode("utf8")
    if type(challenge) == str:
        challenge = challenge.encode("utf8")
    sig = hmac.new(key, challenge, hashlib.sha256).digest()
    return binascii.b2a_base64(sig).strip()


def derive_scram_credential(
    email: str, password: str, salt: Optional[bytes] = None
) -> Dict:
    """
    Derive WAMP-SCRAM credentials from user email and password. The SCRAM parameters used
    are the following (these are also contained in the returned credentials):

    * kdf ``argon2id-13``
    * time cost ``4096``
    * memory cost ``512``
    * parallelism ``1``

    See `draft-irtf-cfrg-argon2 <https://datatracker.ietf.org/doc/draft-irtf-cfrg-argon2/>`__ and
    `argon2-cffi <https://argon2-cffi.readthedocs.io/en/stable/>`__.

    :param email: User email.
    :param password: User password.
    :param salt: Optional salt to use (must be 16 bytes long). If none is given, compute salt
        from email as ``salt = SHA256(email)[:16]``.
    :return: WAMP-SCRAM credentials. When serialized, the returned credentials can be copy-pasted
        into the ``config.json`` node configuration for a Crossbar.io node.
    """
    assert HAS_ARGON, "missing dependency argon2"
    from argon2.low_level import Type, hash_secret

    # derive salt from email
    if not salt:
        m = hashlib.sha256()
        m.update(email.encode("utf8"))
        salt = m.digest()[:16]
    assert len(salt) == 16

    hash_data = hash_secret(
        secret=password.encode("utf8"),
        salt=salt,
        time_cost=4096,
        memory_cost=512,
        parallelism=1,
        hash_len=32,
        type=Type.ID,
        version=19,
    )
    _, tag, v, params, _, salted_password = hash_data.decode("ascii").split("$")
    assert tag == "argon2id"
    assert (
        v == "v=19"
    )  # argon's version 1.3 is represented as 0x13, which is 19 decimal...
    params = {k: v for k, v in [x.split("=") for x in params.split(",")]}

    salted_password = salted_password.encode("ascii")
    client_key = hmac.new(salted_password, b"Client Key", hashlib.sha256).digest()
    stored_key = hashlib.new("sha256", client_key).digest()
    server_key = hmac.new(salted_password, b"Server Key", hashlib.sha256).digest()

    credential = {
        "kdf": "argon2id-13",
        "memory": int(params["m"]),
        "iterations": int(params["t"]),
        "salt": binascii.b2a_hex(salt).decode("ascii"),
        "stored-key": binascii.b2a_hex(stored_key).decode("ascii"),
        "server-key": binascii.b2a_hex(server_key).decode("ascii"),
    }
    return credential
